from __future__ import annotations

import contextlib
import os
from pathlib import Path
from typing import TYPE_CHECKING, Any, Literal

import pyspark
import pyspark.sql.functions as F
import sqlglot as sg
import sqlglot.expressions as sge
from packaging.version import parse as vparse
from pyspark import SparkConf
from pyspark.sql import SparkSession
from pyspark.sql.types import BooleanType, DoubleType, LongType, StringType

import ibis.backends.sql.compilers as sc
import ibis.common.exceptions as com
import ibis.config
import ibis.expr.operations as ops
import ibis.expr.schema as sch
import ibis.expr.types as ir
from ibis import util
from ibis.backends import CanCreateDatabase, CanListCatalog
from ibis.backends.pyspark.converter import PySparkPandasData
from ibis.backends.pyspark.datatypes import PySparkSchema, PySparkType
from ibis.backends.sql import SQLBackend
from ibis.backends.sql.compilers.base import AlterTable, RenameTable
from ibis.expr.operations.udf import InputType
from ibis.legacy.udf.vectorized import _coerce_to_series

try:
    from pyspark.errors import ParseException
    from pyspark.errors.exceptions.connect import SparkConnectGrpcException
except ImportError:
    from pyspark.sql.utils import ParseException

    # Use a dummy class for when spark connect is not available
    class SparkConnectGrpcException(Exception):
        pass


if TYPE_CHECKING:
    from collections.abc import Mapping, Sequence
    from urllib.parse import ParseResult

    import pandas as pd
    import polars as pl
    import pyarrow as pa
    from pyspark.sql.streaming import StreamingQuery

    from ibis.expr.api import Watermark

PYSPARK_VERSION = vparse(pyspark.__version__)
PYSPARK_LT_34 = PYSPARK_VERSION < vparse("3.4")
PYSPARK_LT_35 = PYSPARK_VERSION < vparse("3.5")
ConnectionMode = Literal["streaming", "batch"]


@F.pandas_udf(returnType=DoubleType(), functionType=F.PandasUDFType.SCALAR)
def unwrap_json_float(s: pd.Series) -> pd.Series:
    import json

    import pandas as pd

    def nullify_type_mismatched_value(raw):
        if pd.isna(raw):
            return None

        value = json.loads(raw)
        # exact type check because we want to distinguish between integer
        # and booleans and bool is a subclass of int
        return value if type(value) in (float, int) else None

    return s.map(nullify_type_mismatched_value)


def unwrap_json(typ):
    import json

    import pandas as pd

    type_mapping = {str: StringType(), int: LongType(), bool: BooleanType()}

    @F.pandas_udf(returnType=type_mapping[typ], functionType=F.PandasUDFType.SCALAR)
    def unwrap(s: pd.Series) -> pd.Series:
        def nullify_type_mismatched_value(raw):
            if pd.isna(raw):
                return None

            value = json.loads(raw)
            # exact type check because we want to distinguish between integer
            # and booleans and bool is a subclass of int
            return value if type(value) is typ else None

        return s.map(nullify_type_mismatched_value)

    return unwrap


def _interval_to_string(interval):
    return f"{interval.op().value} {interval.op().dtype.unit.name.lower()}"


class Backend(SQLBackend, CanListCatalog, CanCreateDatabase):
    name = "pyspark"
    compiler = sc.pyspark.compiler

    class Options(ibis.config.Config):
        """PySpark options.

        Attributes
        ----------
        treat_nan_as_null : bool
            Treat NaNs in floating point expressions as NULL.

        """

        treat_nan_as_null: bool = False

    def _from_url(self, url: ParseResult, **kwargs) -> Backend:
        """Construct a PySpark backend from a URL `url`."""
        conf = SparkConf().setAll(kwargs.items())

        if database := url.path[1:]:
            conf = conf.set("spark.sql.warehouse.dir", str(Path(database).absolute()))

        builder = SparkSession.builder.config(conf=conf)
        session = builder.getOrCreate()
        return self.connect(session, **kwargs)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._cached_dataframes = {}

    def do_connect(
        self,
        session: SparkSession | None = None,
        mode: ConnectionMode = "batch",
        **kwargs,
    ) -> None:
        """Create a PySpark `Backend` for use with Ibis.

        Parameters
        ----------
        session
            A `SparkSession` instance.
        mode
            Can be either "batch" or "streaming". If "batch", every source, sink, and
            query executed within this connection will be interpreted as a batch
            workload. If "streaming", every source, sink, and query executed within
            this connection will be interpreted as a streaming workload.
        kwargs
            Additional keyword arguments used to configure the SparkSession.

        Examples
        --------
        >>> import ibis
        >>> from pyspark.sql import SparkSession
        >>> session = SparkSession.builder.getOrCreate()
        >>> ibis.pyspark.connect(session)
        <ibis.backends.pyspark.Backend at 0x...>

        """
        if session is None:
            from pyspark.sql import SparkSession

            session = SparkSession.builder.getOrCreate()

        if mode not in ("batch", "streaming"):
            raise com.IbisInputError(
                f"Invalid connection mode: {mode}, must be `streaming` or `batch`"
            )
        self._mode = mode

        self._session = session

        # Spark internally stores timestamps as UTC values, and timestamp data
        # that is brought in without a specified time zone is converted as
        # local time to UTC with microsecond resolution.
        # https://spark.apache.org/docs/latest/sql-pyspark-pandas-with-arrow.html#timestamp-with-time-zone-semantics
        self._session.conf.set("spark.sql.session.timeZone", "UTC")

        # Databricks Serverless compute only supports limited properties
        # and any attempt to set unsupported properties will result in an error.
        # https://docs.databricks.com/en/spark/conf.html
        with contextlib.suppress(SparkConnectGrpcException):
            self._session.conf.set("spark.sql.mapKeyDedupPolicy", "LAST_WIN")

        for key, value in kwargs.items():
            self._session.conf.set(key, value)

    @util.experimental
    @classmethod
    def from_connection(
        cls, session: SparkSession, mode: ConnectionMode = "batch", **kwargs
    ) -> Backend:
        """Create a PySpark `Backend` from an existing `SparkSession` instance.

        Parameters
        ----------
        session
            A `SparkSession` instance.
        mode
            Can be either "batch" or "streaming". If "batch", every source, sink, and
            query executed within this connection will be interpreted as a batch
            workload. If "streaming", every source, sink, and query executed within
            this connection will be interpreted as a streaming workload.
        kwargs
            Additional keyword arguments used to configure the SparkSession.
        """
        return ibis.pyspark.connect(session, mode, **kwargs)

    def disconnect(self) -> None:
        self._session.stop()

    def _get_schema_using_query(self, query: str) -> sch.Schema:
        df = self.raw_sql(query)
        struct_dtype = PySparkType.to_ibis(df.schema)
        return sch.Schema(struct_dtype)

    @property
    def mode(self) -> ConnectionMode:
        return self._mode

    @property
    def version(self):
        return pyspark.__version__

    @property
    def current_database(self) -> str:
        [(db,)] = self._session.sql("SELECT CURRENT_DATABASE()").collect()
        return db

    @property
    def current_catalog(self) -> str:
        [(catalog,)] = self._session.sql("SELECT CURRENT_CATALOG()").collect()
        return catalog

    @contextlib.contextmanager
    def _active_catalog_database(self, catalog: str | None, db: str | None):
        if catalog is None and db is None:
            yield
            return
        if catalog is not None and PYSPARK_LT_34:
            raise com.UnsupportedArgumentError(
                "Catalogs are not supported in pyspark < 3.4"
            )
        current_catalog = self.current_catalog
        current_db = self.current_database

        # This little horrible bit of work is to avoid trying to set
        # the `CurrentDatabase` inside of a catalog where we don't have permission
        # to do so.  We can't have the catalog and database context managers work
        # separately because we need to:
        # 1. set catalog
        # 2. set database
        # 3. set catalog to previous
        # 4. set database to previous
        #
        # Unity catalog has special handling for "USE CATALOG" and "USE DATABASE"
        # and also has weird permissioning around using `setCurrentCatalog` and
        # `setCurrentDatabase`.
        #
        # We attempt to use the Unity-specific Spark SQL to set CATALOG and DATABASE
        # and if that causes a parser exception we fall back to using the catalog API.
        v = self.compiler.v
        quoted = self.compiler.quoted
        dialect = self.dialect
        catalog_api = self._session.catalog

        try:
            if catalog is not None:
                catalog_sql = sge.Use(
                    kind=v.CATALOG, this=sg.to_identifier(catalog, quoted=quoted)
                ).sql(dialect)

                try:
                    self.raw_sql(catalog_sql)
                except ParseException:
                    catalog_api.setCurrentCatalog(catalog)

            db_sql = sge.Use(
                kind=v.DATABASE, this=sg.to_identifier(db, quoted=quoted)
            ).sql(dialect)

            try:
                self.raw_sql(db_sql)
            except ParseException:
                catalog_api.setCurrentDatabase(db)
            yield
        finally:
            if catalog is not None:
                catalog_sql = sge.Use(
                    kind=v.CATALOG,
                    this=sg.to_identifier(current_catalog, quoted=quoted),
                ).sql(dialect)
                try:
                    self.raw_sql(catalog_sql)
                except ParseException:
                    catalog_api.setCurrentCatalog(current_catalog)

            db_sql = sge.Use(
                kind=v.DATABASE, this=sg.to_identifier(current_db, quoted=quoted)
            ).sql(dialect)

            try:
                self.raw_sql(db_sql)
            except ParseException:
                catalog_api.setCurrentDatabase(current_db)

    @contextlib.contextmanager
    def _active_catalog(self, name: str | None):
        if name is None or PYSPARK_LT_34:
            yield
            return

        prev_catalog = self.current_catalog
        prev_database = self.current_database

        v = self.compiler.v
        quoted = self.compiler.quoted
        dialect = self.dialect

        catalog_sql = sge.Use(
            kind=v.CATALOG, this=sg.to_identifier(name, quoted=quoted)
        ).sql(dialect)
        catalog_api = self._session.catalog

        try:
            try:
                self.raw_sql(catalog_sql)
            except ParseException:
                catalog_api.setCurrentCatalog(name)
            yield
        finally:
            catalog_sql = sge.Use(
                kind=v.CATALOG, this=sg.to_identifier(prev_catalog, quoted=quoted)
            ).sql(dialect)
            db_sql = sge.Use(
                kind=v.DATABASE, this=sg.to_identifier(prev_database, quoted=quoted)
            ).sql(dialect)

            try:
                self.raw_sql(catalog_sql)
                self.raw_sql(db_sql)
            except ParseException:
                catalog_api.setCurrentCatalog(prev_catalog)
                catalog_api.setCurrentDatabase(prev_database)

    def list_catalogs(self, like: str | None = None) -> list[str]:
        catalogs = [res.catalog for res in self._session.sql("SHOW CATALOGS").collect()]
        return self._filter_with_like(catalogs, like)

    def list_databases(
        self, like: str | None = None, catalog: str | None = None
    ) -> list[str]:
        with self._active_catalog(catalog):
            databases = [
                db.namespace for db in self._session.sql("SHOW DATABASES").collect()
            ]
        return self._filter_with_like(databases, like)

    def list_tables(
        self, like: str | None = None, database: str | None = None
    ) -> list[str]:
        """List the tables in the database.

        Parameters
        ----------
        like
            A pattern to use for listing tables.
        database
            Database to list tables from. Default behavior is to show tables in
            the current catalog and database.

            To specify a table in a separate catalog, you can pass in the
            catalog and database as a string `"catalog.database"`, or as a tuple of
            strings `("catalog", "database")`.
        """
        table_loc = self._to_sqlglot_table(database)
        catalog, db = self._to_catalog_db_tuple(table_loc)
        with self._active_catalog(catalog):
            tables = [
                row.tableName
                for row in self._session.sql(
                    f"SHOW TABLES IN {db or self.current_database}"
                ).collect()
            ]
        return self._filter_with_like(tables, like)

    def _wrap_udf_to_return_pandas(self, func, output_dtype):
        def wrapper(*args):
            return _coerce_to_series(func(*args), output_dtype)

        return wrapper

    def _register_udfs(self, expr: ir.Expr) -> None:
        node = expr.op()
        for udf in node.find(ops.ScalarUDF):
            udf_name = self.compiler.__sql_name__(udf)
            udf_return = PySparkType.from_ibis(udf.dtype)
            if udf.__input_type__ == InputType.PANDAS:
                udf_func = self._wrap_udf_to_return_pandas(udf.__func__, udf.dtype)
                spark_udf = F.pandas_udf(udf_func, udf_return, F.PandasUDFType.SCALAR)
            elif udf.__input_type__ == InputType.PYTHON:
                udf_func = udf.__func__
                spark_udf = F.udf(udf_func, udf_return)
            elif udf.__input_type__ == InputType.PYARROW:
                # raise not implemented error if running on pyspark < 3.5
                if PYSPARK_LT_35:
                    raise NotImplementedError(
                        "pyarrow UDFs are only supported in pyspark >= 3.5"
                    )
                udf_func = udf.__func__
                spark_udf = F.udf(udf_func, udf_return, useArrow=True)
            else:
                # Builtin functions don't need to be registered
                continue
            self._session.udf.register(udf_name, spark_udf)
        for udf in node.find(ops.ElementWiseVectorizedUDF):
            udf_name = self.compiler.__sql_name__(udf)
            udf_func = self._wrap_udf_to_return_pandas(udf.func, udf.return_type)
            udf_return = PySparkType.from_ibis(udf.return_type)
            spark_udf = F.pandas_udf(udf_func, udf_return, F.PandasUDFType.SCALAR)
            self._session.udf.register(udf_name, spark_udf)

        for udf in node.find(ops.ReductionVectorizedUDF):
            udf_name = self.compiler.__sql_name__(udf)
            udf_func = self._wrap_udf_to_return_pandas(udf.func, udf.return_type)
            udf_func = udf.func
            udf_return = PySparkType.from_ibis(udf.return_type)
            spark_udf = F.pandas_udf(udf_func, udf_return, F.PandasUDFType.GROUPED_AGG)
            self._session.udf.register(udf_name, spark_udf)

        for typ in (str, int, bool):
            self._session.udf.register(f"unwrap_json_{typ.__name__}", unwrap_json(typ))
        self._session.udf.register("unwrap_json_float", unwrap_json_float)

    def _register_in_memory_table(self, op: ops.InMemoryTable) -> None:
        schema = PySparkSchema.from_ibis(op.schema)
        df = self._session.createDataFrame(data=op.data.to_frame(), schema=schema)
        df.createOrReplaceTempView(op.name)

    def _finalize_memtable(self, name: str) -> None:
        """No-op, otherwise a deadlock can occur when using Spark Connect."""
        if isinstance(session := self._session, pyspark.sql.SparkSession):
            session.catalog.dropTempView(name)

    @contextlib.contextmanager
    def _safe_raw_sql(self, query: str) -> Any:
        yield self.raw_sql(query)

    def raw_sql(self, query: str | sg.Expression, **kwargs: Any) -> Any:
        with contextlib.suppress(AttributeError):
            query = query.sql(dialect=self.dialect)
        return self._session.sql(query, **kwargs)

    def execute(
        self,
        expr: ir.Expr,
        params: Mapping | None = None,
        limit: str | None = "default",
        **kwargs: Any,
    ) -> Any:
        """Execute an expression."""

        self._run_pre_execute_hooks(expr)
        table = expr.as_table()
        sql = self.compile(table, params=params, limit=limit, **kwargs)

        schema = table.schema()

        with self._safe_raw_sql(sql) as query:
            df = query.toPandas()  # blocks until finished
            result = PySparkPandasData.convert_table(df, schema)
        return expr.__pandas_result__(result)

    def create_database(
        self,
        name: str,
        *,
        catalog: str | None = None,
        path: str | Path | None = None,
        force: bool = False,
    ) -> Any:
        """Create a new Spark database.

        Parameters
        ----------
        name
            Database name
        catalog
            Catalog to create database in (defaults to `current_catalog`)
        path
            Path where to store the database data; otherwise uses Spark default
        force
            Whether to append `IF NOT EXISTS` to the database creation SQL

        """
        if path is not None:
            properties = sge.Properties(
                expressions=[sge.LocationProperty(this=sge.convert(str(path)))]
            )
        else:
            properties = None

        sql = sge.Create(
            kind="DATABASE",
            exist=force,
            this=sg.to_identifier(name, quoted=self.compiler.quoted),
            properties=properties,
        )
        with self._active_catalog(catalog):
            with self._safe_raw_sql(sql):
                pass

    def drop_database(
        self, name: str, *, catalog: str | None = None, force: bool = False
    ) -> Any:
        """Drop a Spark database.

        Parameters
        ----------
        name
            Database name
        catalog
            Catalog containing database to drop (defaults to `current_catalog`)
        force
            If False, Spark throws exception if database is not empty or
            database does not exist

        """
        sql = sge.Drop(
            kind="DATABASE",
            exist=force,
            this=sg.to_identifier(name, quoted=self.compiler.quoted),
            cascade=force,
        )
        with self._active_catalog(catalog):
            with self._safe_raw_sql(sql):
                pass

    def get_schema(
        self,
        table_name: str,
        *,
        catalog: str | None = None,
        database: str | None = None,
    ) -> sch.Schema:
        """Return a Schema object for the indicated table and database.

        Parameters
        ----------
        table_name
            Table name. May be fully qualified
        catalog
            Catalog to use
        database
            Database to use to get the active database.

        Returns
        -------
        Schema
            An ibis schema

        """

        table_loc = self._to_sqlglot_table((catalog, database))
        catalog, db = self._to_catalog_db_tuple(table_loc)
        session = self._session
        with self._active_catalog_database(catalog, db):
            try:
                df = session.table(table_name)
                # this is intentionally included in the try block because when
                # using spark connect, the table-not-found exception coming
                # from the server will *NOT* be raised until the schema
                # property is accessed
                struct = PySparkType.to_ibis(df.schema)
            except Exception as e:
                if not session.catalog.tableExists(table_name):
                    raise com.TableNotFound(table_name) from e
                raise

        return sch.Schema(struct)

    def create_table(
        self,
        name: str,
        obj: (
            ir.Table | pd.DataFrame | pa.Table | pl.DataFrame | pl.LazyFrame | None
        ) = None,
        *,
        schema: sch.SchemaLike | None = None,
        database: str | None = None,
        temp: bool | None = None,
        overwrite: bool = False,
        format: str = "parquet",
    ) -> ir.Table:
        """Create a new table in Spark.

        Parameters
        ----------
        name
            Name of the new table.
        obj
            If passed, creates table from `SELECT` statement results
        schema
            Mutually exclusive with `obj`, creates an empty table with a schema
        database
            Database name

            To specify a table in a separate catalog, you can pass in the
            catalog and database as a string `"catalog.database"`, or as a tuple of
            strings `("catalog", "database")`.
        temp
            Whether the new table is temporary (unsupported)
        overwrite
            If `True`, overwrite existing data
        format
            Format of the table on disk

        Returns
        -------
        Table
            The newly created table.

        Examples
        --------
        >>> con.create_table("new_table_name", table_expr)  # quartodoc: +SKIP # doctest: +SKIP

        """
        if temp is True:
            raise NotImplementedError(
                "PySpark backend does not yet support temporary tables"
            )

        table_loc = self._to_sqlglot_table(database)
        catalog, db = self._to_catalog_db_tuple(table_loc)

        if obj is not None:
            if isinstance(obj, ir.Expr):
                table = obj
            else:
                table = ibis.memtable(obj)
            query = self.compile(table)
            mode = "overwrite" if overwrite else "error"
            with self._active_catalog_database(catalog, db):
                self._run_pre_execute_hooks(table)
                df = self._session.sql(query)
                df.write.saveAsTable(name, format=format, mode=mode)
        elif schema is not None:
            schema = ibis.schema(schema)
            schema = PySparkSchema.from_ibis(schema)
            with self._active_catalog_database(catalog, db):
                self._session.catalog.createTable(name, schema=schema, format=format)
        else:
            raise com.IbisError("The schema or obj parameter is required")

        return self.table(name, database=(catalog, db))

    def create_view(
        self,
        name: str,
        obj: ir.Table,
        *,
        database: str | None = None,
        overwrite: bool = False,
    ) -> ir.Table:
        """Create a temporary Spark view from a table expression.

        Parameters
        ----------
        name
            View name
        obj
            Expression to use for the view
        database
            Database name
        overwrite
            Replace an existing view of the same name if it exists

        Returns
        -------
        Table
            The created view

        """
        src = sge.Create(
            this=sg.table(name, db=database, quoted=self.compiler.quoted),
            kind="TEMPORARY VIEW",
            replace=overwrite,
            expression=self.compile(obj),
        )
        self._register_in_memory_tables(obj)
        with self._safe_raw_sql(src):
            pass
        return self.table(name, database=database)

    def rename_table(self, old_name: str, new_name: str) -> None:
        """Rename an existing table.

        Parameters
        ----------
        old_name
            The old name of the table.
        new_name
            The new name of the table.

        """
        quoted = self.compiler.quoted
        old = sge.to_table(old_name, quoted=quoted)
        new = sge.to_table(new_name, quoted=quoted)
        query = AlterTable(
            this=old, exists=False, actions=[RenameTable(this=new, exists=True)]
        )
        with self._safe_raw_sql(query):
            pass

    def compute_stats(
        self,
        name: str,
        database: str | None = None,
        noscan: bool = False,
    ) -> Any:
        """Issue a `COMPUTE STATISTICS` command for a given table.

        Parameters
        ----------
        name
            Table name
        database
            Database name
        noscan
            If `True`, collect only basic statistics for the table (number of
            rows, size in bytes).

        """
        maybe_noscan = " NOSCAN" * noscan
        table = sg.table(name, db=database, quoted=self.compiler.quoted).sql(
            dialect=self.dialect
        )
        return self.raw_sql(f"ANALYZE TABLE {table} COMPUTE STATISTICS{maybe_noscan}")

    def _create_cached_table(self, name, expr):
        query = self.compile(expr)
        t = self._session.sql(query).cache()
        assert t.is_cached
        t.createTempView(name)
        # store the underlying spark dataframe so we can release memory when
        # asked to, instead of when the session ends
        self._cached_dataframes[name] = t
        return self.table(name)

    def _drop_cached_table(self, name):
        self._session.catalog.dropTempView(name)
        t = self._cached_dataframes.pop(name)
        t.unpersist()
        assert not t.is_cached

    def read_delta(
        self,
        path: str | Path,
        table_name: str | None = None,
        **kwargs: Any,
    ) -> ir.Table:
        """Register a Delta Lake table as a table in the current database.

        Parameters
        ----------
        path
            The path to the Delta Lake table.
        table_name
            An optional name to use for the created table. This defaults to
            a random generated name.
        kwargs
            Additional keyword arguments passed to PySpark.
            https://spark.apache.org/docs/latest/api/python/reference/pyspark.sql/api/pyspark.sql.DataFrameReader.load.html

        Returns
        -------
        ir.Table
            The just-registered table

        """
        if self.mode == "streaming":
            raise NotImplementedError(
                "Reading a Delta Lake table in streaming mode is not supported"
            )
        path = util.normalize_filename(path)
        spark_df = self._session.read.format("delta").load(path, **kwargs)
        table_name = table_name or util.gen_name("read_delta")

        spark_df.createOrReplaceTempView(table_name)
        return self.table(table_name)

    def read_parquet(
        self,
        path: str | Path,
        table_name: str | None = None,
        **kwargs: Any,
    ) -> ir.Table:
        """Register a parquet file as a table in the current database.

        Parameters
        ----------
        path
            The data source. May be a path to a file or directory of parquet files.
        table_name
            An optional name to use for the created table. This defaults to
            a random generated name.
        kwargs
            Additional keyword arguments passed to PySpark.
            https://spark.apache.org/docs/latest/api/python/reference/pyspark.sql/api/pyspark.sql.DataFrameReader.parquet.html

        Returns
        -------
        ir.Table
            The just-registered table

        """
        if self.mode == "streaming":
            raise NotImplementedError(
                "Pyspark in streaming mode does not support direction registration of parquet files. "
                "Please use `read_parquet_dir` instead."
            )
        path = util.normalize_filename(path)
        spark_df = self._session.read.parquet(path, **kwargs)
        table_name = table_name or util.gen_name("read_parquet")

        spark_df.createOrReplaceTempView(table_name)
        return self.table(table_name)

    def read_csv(
        self,
        source_list: str | list[str] | tuple[str],
        table_name: str | None = None,
        **kwargs: Any,
    ) -> ir.Table:
        """Register a CSV file as a table in the current database.

        Parameters
        ----------
        source_list
            The data source(s). May be a path to a file or directory of CSV files, or an
            iterable of CSV files.
        table_name
            An optional name to use for the created table. This defaults to
            a random generated name.
        kwargs
            Additional keyword arguments passed to PySpark loading function.
            https://spark.apache.org/docs/latest/api/python/reference/pyspark.sql/api/pyspark.sql.DataFrameReader.csv.html

        Returns
        -------
        ir.Table
            The just-registered table

        """
        if self.mode == "streaming":
            raise NotImplementedError(
                "Pyspark in streaming mode does not support direction registration of CSV files. "
                "Please use `read_csv_dir` instead."
            )
        inferSchema = kwargs.pop("inferSchema", True)
        header = kwargs.pop("header", True)
        source_list = util.normalize_filenames(source_list)
        spark_df = self._session.read.csv(
            source_list, inferSchema=inferSchema, header=header, **kwargs
        )
        table_name = table_name or util.gen_name("read_csv")

        spark_df.createOrReplaceTempView(table_name)
        return self.table(table_name)

    def read_json(
        self,
        source_list: str | Sequence[str],
        table_name: str | None = None,
        **kwargs: Any,
    ) -> ir.Table:
        """Register a JSON file as a table in the current database.

        Parameters
        ----------
        source_list
            The data source(s). May be a path to a file or directory of JSON files, or an
            iterable of JSON files.
        table_name
            An optional name to use for the created table. This defaults to
            a random generated name.
        kwargs
            Additional keyword arguments passed to PySpark loading function.
            https://spark.apache.org/docs/latest/api/python/reference/pyspark.sql/api/pyspark.sql.DataFrameReader.json.html

        Returns
        -------
        ir.Table
            The just-registered table

        """
        if self.mode == "streaming":
            raise NotImplementedError(
                "Pyspark in streaming mode does not support direction registration of JSON files. "
                "Please use `read_json_dir` instead."
            )
        source_list = util.normalize_filenames(source_list)
        spark_df = self._session.read.json(source_list, **kwargs)
        table_name = table_name or util.gen_name("read_json")

        spark_df.createOrReplaceTempView(table_name)
        return self.table(table_name)

    @util.experimental
    def to_delta(
        self,
        expr: ir.Table,
        path: str | Path,
        params: Mapping[ir.Scalar, Any] | None = None,
        limit: int | str | None = None,
        **kwargs: Any,
    ) -> None:
        """Write the results of executing the given expression to a Delta Lake table.

        This method is eager and will execute the associated expression
        immediately.

        Parameters
        ----------
        expr
            The ibis expression to execute and persist to a Delta Lake table.
        path
            The data source. A string or Path to the Delta Lake table.
        params
            Mapping of scalar parameter expressions to value.
        limit
            An integer to effect a specific row limit. A value of `None` means
            "no limit". The default is in `ibis/config.py`.
        **kwargs
            Additional keyword arguments passed to
            [pyspark.sql.DataFrameWriter](https://spark.apache.org/docs/latest/api/python/reference/pyspark.sql/api/pyspark.sql.DataFrameWriter.html).

        """
        if self.mode == "streaming":
            raise NotImplementedError(
                "Writing to a Delta Lake table in streaming mode is not supported"
            )
        self._run_pre_execute_hooks(expr)
        df = self._session.sql(self.compile(expr, params=params, limit=limit))
        df.write.format("delta").save(os.fspath(path), **kwargs)

    def to_pyarrow(
        self,
        expr: ir.Expr,
        params: Mapping[ir.Scalar, Any] | None = None,
        limit: int | str | None = None,
        **kwargs: Any,
    ) -> pa.Table:
        if self.mode == "streaming":
            raise NotImplementedError(
                "PySpark in streaming mode does not support to_pyarrow"
            )
        import pyarrow as pa
        import pyarrow_hotfix  # noqa: F401

        from ibis.formats.pyarrow import PyArrowData

        table_expr = expr.as_table()
        output = pa.Table.from_pandas(
            self.execute(table_expr, params=params, limit=limit, **kwargs),
            preserve_index=False,
        )
        table = PyArrowData.convert_table(output, table_expr.schema())
        return expr.__pyarrow_result__(table)

    def to_pyarrow_batches(
        self,
        expr: ir.Expr,
        *,
        params: Mapping[ir.Scalar, Any] | None = None,
        limit: int | str | None = None,
        chunk_size: int = 1000000,
        **kwargs: Any,
    ) -> pa.ipc.RecordBatchReader:
        if self.mode == "streaming":
            raise NotImplementedError(
                "PySpark in streaming mode does not support to_pyarrow_batches"
            )
        pa = self._import_pyarrow()
        pa_table = self.to_pyarrow(
            expr.as_table(), params=params, limit=limit, **kwargs
        )
        return pa.ipc.RecordBatchReader.from_batches(
            pa_table.schema, pa_table.to_batches(max_chunksize=chunk_size)
        )

    @util.experimental
    def read_kafka(
        self,
        table_name: str | None = None,
        *,
        watermark: Watermark | None = None,
        auto_parse: bool = False,
        schema: sch.Schema | None = None,
        options: Mapping[str, str] | None = None,
    ) -> ir.Table:
        """Register a Kafka topic as a table.

        Parameters
        ----------
        table_name
            An optional name to use for the created table. This defaults to
            a sequentially generated name.
        watermark
            Watermark strategy for the table.
        auto_parse
            Whether to parse Kafka messages automatically. If `False`, the source is read
            as binary keys and values. If `True`, the key is discarded and the value is
            parsed using the provided schema.
        schema
            Schema of the value of the Kafka messages.
        options
            Additional arguments passed to PySpark as .option("key", "value").
            https://spark.apache.org/docs/latest/structured-streaming-kafka-integration.html

        Returns
        -------
        ir.Table
            The just-registered table
        """
        if self.mode == "batch":
            raise NotImplementedError(
                "Reading from Kafka in batch mode is not supported"
            )
        spark_df = self._session.readStream.format("kafka")
        for k, v in (options or {}).items():
            spark_df = spark_df.option(k, v)
        spark_df = spark_df.load()

        # parse the values of the Kafka messages using the provided schema
        if auto_parse:
            if schema is None:
                raise com.IbisError(
                    "When auto_parse is True, a schema must be provided to parse the messages"
                )
            schema = PySparkSchema.from_ibis(schema)
            spark_df = spark_df.select(
                F.from_json(F.col("value").cast("string"), schema).alias("parsed_value")
            ).select("parsed_value.*")

        if watermark is not None:
            spark_df = spark_df.withWatermark(
                watermark.time_col,
                _interval_to_string(watermark.allowed_delay),
            )

        table_name = table_name or util.gen_name("read_kafka")
        spark_df.createOrReplaceTempView(table_name)
        return self.table(table_name)

    @util.experimental
    def to_kafka(
        self,
        expr: ir.Expr,
        *,
        auto_format: bool = False,
        options: Mapping[str, str] | None = None,
        params: Mapping | None = None,
        limit: str | None = "default",
    ) -> StreamingQuery:
        """Write the results of executing the given expression to a Kafka topic.

        This method does not return outputs. Streaming queries are run continuously in
        the background.

        Parameters
        ----------
        expr
            The ibis expression to execute and persist to a Kafka topic.
        auto_format
            Whether to format the Kafka messages before writing. If `False`, the output is
            written as-is. If `True`, the output is converted into JSON and written as the
            value of the Kafka messages.
        options
            PySpark Kafka write arguments.
            https://spark.apache.org/docs/latest/structured-streaming-kafka-integration.html
        params
            Mapping of scalar parameter expressions to value.
        limit
            An integer to effect a specific row limit. A value of `None` means
            "no limit". The default is in `ibis/config.py`.

        Returns
        -------
        StreamingQuery
            A Pyspark StreamingQuery object
        """
        if self.mode == "batch":
            raise NotImplementedError("Writing to Kafka in batch mode is not supported")
        df = self._session.sql(self.compile(expr, params=params, limit=limit))
        if auto_format:
            df = df.select(
                F.to_json(F.struct([F.col(c).alias(c) for c in df.columns])).alias(
                    "value"
                )
            )
        sq = df.writeStream.format("kafka")
        for k, v in (options or {}).items():
            sq = sq.option(k, v)
        sq.start()
        return sq

    @util.experimental
    def read_csv_dir(
        self,
        path: str | Path,
        table_name: str | None = None,
        watermark: Watermark | None = None,
        **kwargs: Any,
    ) -> ir.Table:
        """Register a CSV directory as a table in the current database.

        Parameters
        ----------
        path
            The data source.
        table_name
            An optional name to use for the created table. This defaults to
            a random generated name.
        watermark
            Watermark strategy for the table.
        kwargs
            Additional keyword arguments passed to PySpark loading function.
            https://spark.apache.org/docs/latest/api/python/reference/pyspark.ss/api/pyspark.sql.streaming.DataStreamReader.csv.html

        Returns
        -------
        ir.Table
            The just-registered table

        """
        inferSchema = kwargs.pop("inferSchema", True)
        header = kwargs.pop("header", True)
        path = util.normalize_filename(path)
        if self.mode == "batch":
            spark_df = self._session.read.csv(
                path, inferSchema=inferSchema, header=header, **kwargs
            )
            if watermark is not None:
                raise com.IbisInputError("Watermark is not supported in batch mode")
        elif self.mode == "streaming":
            spark_df = self._session.readStream.csv(
                path, inferSchema=inferSchema, header=header, **kwargs
            )
            if watermark is not None:
                spark_df = spark_df.withWatermark(
                    watermark.time_col,
                    _interval_to_string(watermark.allowed_delay),
                )
        table_name = table_name or util.gen_name("read_csv_dir")

        spark_df.createOrReplaceTempView(table_name)
        return self.table(table_name)

    @util.experimental
    def read_parquet_dir(
        self,
        path: str | Path,
        table_name: str | None = None,
        watermark: Watermark | None = None,
        schema: sch.Schema | None = None,
        **kwargs: Any,
    ) -> ir.Table:
        """Register a parquet file as a table in the current database.

        Parameters
        ----------
        path
            The data source. A directory of parquet files.
        table_name
            An optional name to use for the created table. This defaults to
            a random generated name.
        watermark
            Watermark strategy for the table.
        schema
            Schema of the parquet source.
        kwargs
            Additional keyword arguments passed to PySpark.
            https://spark.apache.org/docs/latest/api/python/reference/pyspark.ss/api/pyspark.sql.streaming.DataStreamReader.parquet.html

        Returns
        -------
        ir.Table
            The just-registered table

        """
        path = util.normalize_filename(path)
        if self.mode == "batch":
            spark_df = self._session.read
            if schema is not None:
                spark_df = spark_df.schema(PySparkSchema.from_ibis(schema))
            spark_df = spark_df.parquet(path, **kwargs)
            if watermark is not None:
                raise com.IbisInputError("Watermark is not supported in batch mode")
        elif self.mode == "streaming":
            spark_df = self._session.readStream
            if schema is not None:
                spark_df = spark_df.schema(PySparkSchema.from_ibis(schema))
            spark_df = spark_df.parquet(path, **kwargs)
            if watermark is not None:
                spark_df = spark_df.withWatermark(
                    watermark.time_col,
                    _interval_to_string(watermark.allowed_delay),
                )
        table_name = table_name or util.gen_name("read_parquet_dir")

        spark_df.createOrReplaceTempView(table_name)
        return self.table(table_name)

    @util.experimental
    def read_json_dir(
        self,
        path: str | Path,
        table_name: str | None = None,
        watermark: Watermark | None = None,
        **kwargs: Any,
    ) -> ir.Table:
        """Register a JSON file as a table in the current database.

        Parameters
        ----------
        path
            The data source. A directory of JSON files.
        table_name
            An optional name to use for the created table. This defaults to
            a random generated name.
        watermark
            Watermark strategy for the table.
        kwargs
            Additional keyword arguments passed to PySpark loading function.
            https://spark.apache.org/docs/latest/api/python/reference/pyspark.ss/api/pyspark.sql.streaming.DataStreamReader.json.html

        Returns
        -------
        ir.Table
            The just-registered table

        """
        path = util.normalize_filename(path)
        if self.mode == "batch":
            spark_df = self._session.read.json(path, **kwargs)
            if watermark is not None:
                raise com.IbisInputError("Watermark is not supported in batch mode")
        elif self.mode == "streaming":
            spark_df = self._session.readStream.json(path, **kwargs)
            if watermark is not None:
                spark_df = spark_df.withWatermark(
                    watermark.time_col,
                    _interval_to_string(watermark.allowed_delay),
                )
        table_name = table_name or util.gen_name("read_json_dir")

        spark_df.createOrReplaceTempView(table_name)
        return self.table(table_name)

    def _to_filesystem_output(
        self,
        expr: ir.Expr,
        format: str,
        path: str | Path,
        params: Mapping[ir.Scalar, Any] | None = None,
        limit: int | str | None = None,
        options: Mapping[str, str] | None = None,
    ) -> StreamingQuery | None:
        df = self._session.sql(self.compile(expr, params=params, limit=limit))
        if self.mode == "batch":
            df = df.write.format(format)
            for k, v in (options or {}).items():
                df = df.option(k, v)
            df.save(os.fspath(path))
            return None
        sq = df.writeStream.format(format)
        sq = sq.option("path", os.fspath(path))
        for k, v in (options or {}).items():
            sq = sq.option(k, v)
        sq.start()
        return sq

    @util.experimental
    def to_parquet_dir(
        self,
        expr: ir.Expr,
        path: str | Path,
        params: Mapping[ir.Scalar, Any] | None = None,
        limit: int | str | None = None,
        options: Mapping[str, str] | None = None,
    ) -> StreamingQuery | None:
        """Write the results of executing the given expression to a parquet directory.

        Parameters
        ----------
        expr
            The ibis expression to execute and persist to parquet.
        path
            The data source. A string or Path to the parquet directory.
        params
            Mapping of scalar parameter expressions to value.
        limit
            An integer to effect a specific row limit. A value of `None` means
            "no limit". The default is in `ibis/config.py`.
        options
            Additional keyword arguments passed to pyspark.sql.streaming.DataStreamWriter

        Returns
        -------
        StreamingQuery | None
            Returns a Pyspark StreamingQuery object if in streaming mode, otherwise None
        """
        self._run_pre_execute_hooks(expr)
        return self._to_filesystem_output(expr, "parquet", path, params, limit, options)

    @util.experimental
    def to_csv_dir(
        self,
        expr: ir.Expr,
        path: str | Path,
        params: Mapping[ir.Scalar, Any] | None = None,
        limit: int | str | None = None,
        options: Mapping[str, str] | None = None,
    ) -> StreamingQuery | None:
        """Write the results of executing the given expression to a CSV directory.

        Parameters
        ----------
        expr
            The ibis expression to execute and persist to CSV.
        path
            The data source. A string or Path to the CSV directory.
        params
            Mapping of scalar parameter expressions to value.
        limit
            An integer to effect a specific row limit. A value of `None` means
            "no limit". The default is in `ibis/config.py`.
        options
            Additional keyword arguments passed to pyspark.sql.streaming.DataStreamWriter

        Returns
        -------
        StreamingQuery | None
            Returns a Pyspark StreamingQuery object if in streaming mode, otherwise None
        """
        self._run_pre_execute_hooks(expr)
        return self._to_filesystem_output(expr, "csv", path, params, limit, options)



from __future__ import annotations

import ibis
from ibis import _


def test_more_than_one_quantile(snapshot):
    tables = ibis.table(name="t", schema={"ROW_COUNT": "int"})

    expr = tables.aggregate(
        quantile_0_25=_.ROW_COUNT.quantile(0.25),
        quantile_0_75=_.ROW_COUNT.quantile(0.75),
    )

    sql = ibis.to_sql(expr, dialect="snowflake")
    snapshot.assert_match(sql, "two_quantiles.sql")



import ibis


t = ibis.table(name="t", schema={"a": "int64", "b": "string"})
f = t.filter((t.b == "m"))
agg = f.aggregate([f.a.sum().name("sum"), f.a.max()], by=[f.b])
f1 = agg.filter((agg["Max(a)"] == 2))

result = f1.select(f1.b, f1.sum)



from __future__ import annotations

import pytest


@pytest.fixture(scope="session")
def call_center(backend):
    return backend.ds("call_center")


@pytest.fixture(scope="session")
def catalog_page(backend):
    return backend.ds("catalog_page")


@pytest.fixture(scope="session")
def catalog_returns(backend):
    return backend.ds("catalog_returns")


@pytest.fixture(scope="session")
def catalog_sales(backend):
    return backend.ds("catalog_sales")


@pytest.fixture(scope="session")
def customer(backend):
    return backend.ds("customer")


@pytest.fixture(scope="session")
def customer_address(backend):
    return backend.ds("customer_address")


@pytest.fixture(scope="session")
def customer_demographics(backend):
    return backend.ds("customer_demographics")


@pytest.fixture(scope="session")
def date_dim(backend):
    return backend.ds("date_dim")


@pytest.fixture(scope="session")
def household_demographics(backend):
    return backend.ds("household_demographics")


@pytest.fixture(scope="session")
def income_band(backend):
    return backend.ds("income_band")


@pytest.fixture(scope="session")
def inventory(backend):
    return backend.ds("inventory")


@pytest.fixture(scope="session")
def item(backend):
    return backend.ds("item")


@pytest.fixture(scope="session")
def promotion(backend):
    return backend.ds("promotion")


@pytest.fixture(scope="session")
def reason(backend):
    return backend.ds("reason")


@pytest.fixture(scope="session")
def ship_mode(backend):
    return backend.ds("ship_mode")


@pytest.fixture(scope="session")
def store(backend):
    return backend.ds("store")


@pytest.fixture(scope="session")
def store_returns(backend):
    return backend.ds("store_returns")


@pytest.fixture(scope="session")
def store_sales(backend):
    return backend.ds("store_sales")


@pytest.fixture(scope="session")
def time_dim(backend):
    return backend.ds("time_dim")


@pytest.fixture(scope="session")
def warehouse(backend):
    return backend.ds("warehouse")


@pytest.fixture(scope="session")
def web_page(backend):
    return backend.ds("web_page")


@pytest.fixture(scope="session")
def web_returns(backend):
    return backend.ds("web_returns")


@pytest.fixture(scope="session")
def web_sales(backend):
    return backend.ds("web_sales")


@pytest.fixture(scope="session")
def web_site(backend):
    return backend.ds("web_site")



from __future__ import annotations

import pytest
from pytest import param

import ibis
import ibis.common.exceptions as com
import ibis.expr.datatypes as dt
from ibis.backends.tests.errors import PySparkPythonException
from ibis.conftest import IS_SPARK_REMOTE
from ibis.legacy.udf.vectorized import analytic, elementwise, reduction

np = pytest.importorskip("numpy")
pd = pytest.importorskip("pandas")

pytestmark = [
    pytest.mark.notimpl(["druid", "oracle", "risingwave"]),
    pytest.mark.notyet(
        ["pyspark"],
        condition=IS_SPARK_REMOTE,
        raises=PySparkPythonException,
        # TODO(cpcloud): this API is deprecated in 10.0.0, no use copypasting a
        # bunch of markers just for two passing tests
        strict=False,
        reason="remote udfs not yet tested due to environment complexities",
    ),
]


def _format_udf_return_type(func, result_formatter):
    """Call the given udf and return its result according to the given format
    (e.g. in the form of a list, pd.Series, np.array, etc.)"""

    def _wrapper(*args, **kwargs):
        result = func(*args, **kwargs)
        return result_formatter(result)

    return _wrapper


def _format_struct_udf_return_type(func, result_formatter):
    """Call the given struct udf and return its result according to the given
    format (e.g. in the form of a list, pd.Series, np.array, etc.)"""

    def _wrapper(*args, **kwargs):
        result = func(*args, **kwargs)
        return result_formatter(*result)

    return _wrapper


# elementwise UDF
def add_one(s):
    assert isinstance(s, pd.Series), type(s)
    return s + 1


def create_add_one_udf(result_formatter, id):
    with pytest.warns(FutureWarning, match="v9.0"):

        @elementwise(input_type=[dt.double], output_type=dt.double)
        def add_one_legacy(s):
            return result_formatter(add_one(s))

    @ibis.udf.scalar.pandas
    def add_one_udf(s: float) -> float:
        return result_formatter(add_one(s))

    yield param(add_one_legacy, id=f"add_one_legacy_{id}")
    yield param(add_one_udf, id=f"add_one_modern_{id}")


add_one_udfs = [
    *create_add_one_udf(result_formatter=lambda v: v, id="series"),
    *create_add_one_udf(result_formatter=lambda v: np.array(v), id="array"),
    *create_add_one_udf(result_formatter=lambda v: list(v), id="list"),
]


# analytic UDF
def calc_zscore(s):
    assert isinstance(s, pd.Series)
    return (s - s.mean()) / s.std()


def create_calc_zscore_udf(result_formatter):
    with pytest.warns(FutureWarning, match="v9.0"):
        return analytic(input_type=[dt.double], output_type=dt.double)(
            _format_udf_return_type(calc_zscore, result_formatter)
        )


calc_zscore_udfs = [
    create_calc_zscore_udf(result_formatter=lambda v: v),  # pd.Series,
    create_calc_zscore_udf(result_formatter=lambda v: np.array(v)),  # np.array,
    create_calc_zscore_udf(result_formatter=lambda v: list(v)),  # list,
]

with pytest.warns(FutureWarning, match="v9.0"):

    @reduction(input_type=[dt.double], output_type=dt.double)
    def calc_mean(s):
        assert isinstance(s, (np.ndarray, pd.Series))
        return s.mean()


# elementwise multi-column UDF
def add_one_struct(v):
    assert isinstance(v, pd.Series)
    return v + 1, v + 2


def create_add_one_struct_udf(result_formatter):
    with pytest.warns(FutureWarning, match="v9.0"):
        return elementwise(
            input_type=[dt.double],
            output_type=dt.Struct({"col1": dt.double, "col2": dt.double}),
        )(_format_struct_udf_return_type(add_one_struct, result_formatter))


add_one_struct_udfs = [
    param(
        create_add_one_struct_udf(result_formatter=lambda v1, v2: (v1, v2)),
        id="tuple_of_series",
    ),
    param(
        create_add_one_struct_udf(result_formatter=lambda v1, v2: [v1, v2]),
        id="list_of_series",
    ),
    param(
        create_add_one_struct_udf(
            result_formatter=lambda v1, v2: (np.array(v1), np.array(v2))
        ),
        id="tuple_of_ndarray",
    ),
    param(
        create_add_one_struct_udf(
            result_formatter=lambda v1, v2: [np.array(v1), np.array(v2)]
        ),
        id="list_of_ndarray",
    ),
    param(
        create_add_one_struct_udf(
            result_formatter=lambda v1, v2: np.array([np.array(v1), np.array(v2)])
        ),
        id="ndarray_of_ndarray",
    ),
    param(
        create_add_one_struct_udf(
            result_formatter=lambda v1, v2: pd.DataFrame({"col1": v1, "col2": v2})
        ),
        id="dataframe",
    ),
]

with pytest.warns(FutureWarning, match="v9.0"):

    @elementwise(
        input_type=[dt.double],
        output_type=dt.Struct({"double_col": dt.double, "col2": dt.double}),
    )
    def overwrite_struct_elementwise(v):
        assert isinstance(v, pd.Series)
        return v + 1, v + 2

    @elementwise(
        input_type=[dt.double],
        output_type=dt.Struct(
            {"double_col": dt.double, "col2": dt.double, "float_col": dt.double}
        ),
    )
    def multiple_overwrite_struct_elementwise(v):
        assert isinstance(v, pd.Series)
        return v + 1, v + 2, v + 3


with pytest.warns(FutureWarning, match="v9.0"):

    @analytic(
        input_type=[dt.double, dt.double],
        output_type=dt.Struct({"double_col": dt.double, "demean_weight": dt.double}),
    )
    def overwrite_struct_analytic(v, w):
        assert isinstance(v, pd.Series)
        assert isinstance(w, pd.Series)
        return v - v.mean(), w - w.mean()


# analytic multi-column UDF
def demean_struct(v, w):
    assert isinstance(v, pd.Series)
    assert isinstance(w, pd.Series)
    return v - v.mean(), w - w.mean()


def create_demean_struct_udf(result_formatter):
    with pytest.warns(FutureWarning, match="v9.0"):
        return analytic(
            input_type=[dt.double, dt.double],
            output_type=dt.Struct({"demean": dt.double, "demean_weight": dt.double}),
        )(_format_struct_udf_return_type(demean_struct, result_formatter))


demean_struct_udfs = [
    create_demean_struct_udf(
        result_formatter=lambda v1, v2: (v1, v2)
    ),  # tuple of pd.Series,
    create_demean_struct_udf(
        result_formatter=lambda v1, v2: [v1, v2]
    ),  # list of pd.Series,
    create_demean_struct_udf(
        result_formatter=lambda v1, v2: (np.array(v1), np.array(v2))
    ),  # tuple of np.array,
    create_demean_struct_udf(
        result_formatter=lambda v1, v2: [np.array(v1), np.array(v2)]
    ),  # list of np.array,
    create_demean_struct_udf(
        result_formatter=lambda v1, v2: np.array([np.array(v1), np.array(v2)])
    ),  # np.array of np.array,
    create_demean_struct_udf(
        result_formatter=lambda v1, v2: pd.DataFrame(
            {"demean": v1, "demean_weight": v2}
        )
    ),  # pd.DataFrame,
]


# reduction multi-column UDF
def mean_struct(v, w):
    assert isinstance(v, (np.ndarray, pd.Series))
    assert isinstance(w, (np.ndarray, pd.Series))
    return v.mean(), w.mean()


def create_mean_struct_udf(result_formatter):
    with pytest.warns(FutureWarning, match="v9.0"):
        return reduction(
            input_type=[dt.double, dt.int64],
            output_type=dt.Struct({"mean": dt.double, "mean_weight": dt.double}),
        )(_format_struct_udf_return_type(mean_struct, result_formatter))


mean_struct_udfs = [
    create_mean_struct_udf(
        result_formatter=lambda v1, v2: (v1, v2)
    ),  # tuple of scalar,
    create_mean_struct_udf(result_formatter=lambda v1, v2: [v1, v2]),  # list of scalar,
    create_mean_struct_udf(
        result_formatter=lambda v1, v2: np.array([v1, v2])
    ),  # np.array of scalar
]

with pytest.warns(FutureWarning, match="v9.0"):

    @reduction(
        input_type=[dt.double, dt.int64],
        output_type=dt.Struct({"double_col": dt.double, "mean_weight": dt.double}),
    )
    def overwrite_struct_reduction(v, w):
        assert isinstance(v, (np.ndarray, pd.Series))
        assert isinstance(w, (np.ndarray, pd.Series))
        return v.mean(), w.mean()

    @reduction(
        input_type=[dt.double],
        output_type=dt.Array(dt.double),
    )
    def quantiles(series, *, quantiles):
        return series.quantile(quantiles)


@pytest.mark.parametrize(
    "udf", create_add_one_udf(result_formatter=lambda v: v, id="series")
)
def test_elementwise_udf(udf_backend, udf_alltypes, udf_df, udf):
    expr = udf(udf_alltypes["double_col"])
    result = expr.execute()

    expected_func = getattr(expr.op(), "__func__", getattr(udf, "func", None))
    assert (
        expected_func is not None
    ), f"neither __func__ nor func attributes found on {udf} or expr object"

    expected = expected_func(udf_df["double_col"])
    udf_backend.assert_series_equal(result, expected, check_names=False)


@pytest.mark.parametrize("udf", add_one_udfs)
def test_elementwise_udf_mutate(udf_backend, udf_alltypes, udf_df, udf):
    udf_expr = udf(udf_alltypes["double_col"])
    expr = udf_alltypes.mutate(incremented=udf_expr)
    result = expr.execute()

    expected_func = getattr(udf_expr.op(), "__func__", getattr(udf, "func", None))
    assert (
        expected_func is not None
    ), f"neither __func__ nor func attributes found on {udf} or expr object"

    expected = udf_df.assign(incremented=expected_func(udf_df["double_col"]))
    udf_backend.assert_series_equal(result["incremented"], expected["incremented"])


@pytest.mark.notimpl(["pyspark"])
def test_analytic_udf(udf_backend, udf_alltypes, udf_df):
    calc_zscore_udf = create_calc_zscore_udf(result_formatter=lambda v: v)
    result = calc_zscore_udf(udf_alltypes["double_col"]).execute()
    expected = calc_zscore_udf.func(udf_df["double_col"])
    udf_backend.assert_series_equal(result, expected, check_names=False)


@pytest.mark.parametrize("udf", calc_zscore_udfs)
@pytest.mark.notimpl(["pyspark"])
def test_analytic_udf_mutate(udf_backend, udf_alltypes, udf_df, udf):
    expr = udf_alltypes.mutate(zscore=udf(udf_alltypes["double_col"]))
    result = expr.execute()

    expected = udf_df.assign(zscore=udf.func(udf_df["double_col"]))

    udf_backend.assert_series_equal(result["zscore"], expected["zscore"])


def test_reduction_udf(udf_alltypes, udf_df):
    result = calc_mean(udf_alltypes["double_col"]).execute()
    expected = udf_df["double_col"].mean()
    assert result == expected


def test_reduction_udf_array_return_type(udf_backend, udf_alltypes, udf_df):
    """Tests reduction UDF returning an array."""
    qs = [0.25, 0.75]
    expr = udf_alltypes.mutate(q=quantiles(udf_alltypes["int_col"], quantiles=qs))
    result = expr.execute()

    expected = udf_df.assign(
        q=pd.Series(
            [quantiles.func(udf_df["int_col"], quantiles=qs).tolist()] * len(udf_df)
        )
    )
    udf_backend.assert_frame_equal(result, expected)


def test_reduction_udf_on_empty_data(udf_backend, udf_alltypes):
    """Test that summarization can handle empty data."""
    # First filter down to zero rows
    t = udf_alltypes.filter(udf_alltypes["int_col"] > np.inf)
    result = t.group_by("year").aggregate(mean=calc_mean(t["int_col"])).execute()
    expected = pd.DataFrame({"year": [], "mean": []})
    # We check that the result is an empty DataFrame,
    # rather than an error.
    udf_backend.assert_frame_equal(result, expected, check_dtype=False)


def test_output_type_in_list_invalid():
    # Test that an error is raised if UDF output type is wrapped in a list

    with pytest.raises(
        com.IbisTypeError,
        match="The output type of a UDF must be a single datatype.",
    ):
        with pytest.warns(FutureWarning, match="v9.0"):

            @elementwise(input_type=[dt.double], output_type=[dt.double])
            def _(s):
                return s + 1


def test_valid_kwargs(udf_backend, udf_alltypes, udf_df):
    # Test different forms of UDF definition with keyword arguments
    with pytest.warns(FutureWarning, match="v9.0"):

        @elementwise(input_type=[dt.double], output_type=dt.double)
        def foo1(v):
            # Basic UDF with kwargs
            return v + 1

        @elementwise(input_type=[dt.double], output_type=dt.double)
        def foo2(v, *, amount):
            # UDF with keyword only arguments
            return v + amount

        @elementwise(input_type=[dt.double], output_type=dt.double)
        def foo3(v, **kwargs):
            # UDF with kwargs
            return v + kwargs.get("amount", 1)

    expr = udf_alltypes.mutate(
        v1=foo1(udf_alltypes["double_col"]),
        v2=foo2(udf_alltypes["double_col"], amount=1),
        v3=foo2(udf_alltypes["double_col"], amount=2),
        v4=foo3(udf_alltypes["double_col"]),
        v5=foo3(udf_alltypes["double_col"], amount=2),
        v6=foo3(udf_alltypes["double_col"], amount=3),
    )
    result = expr.execute()

    expected = udf_df.assign(
        v1=udf_df["double_col"] + 1,
        v2=udf_df["double_col"] + 1,
        v3=udf_df["double_col"] + 2,
        v4=udf_df["double_col"] + 1,
        v5=udf_df["double_col"] + 2,
        v6=udf_df["double_col"] + 3,
    )

    udf_backend.assert_frame_equal(result, expected)


def test_valid_args(udf_backend, udf_alltypes, udf_df):
    # Test different forms of UDF definition with *args
    with pytest.warns(FutureWarning, match="v9.0"):

        @elementwise(input_type=[dt.double, dt.int32], output_type=dt.double)
        def foo1(*args):
            return args[0] + args[1]

        @elementwise(input_type=[dt.double, dt.int32], output_type=dt.double)
        def foo2(v, *args):
            return v + args[0]

    result = udf_alltypes.mutate(
        v1=foo1(udf_alltypes["double_col"], udf_alltypes["int_col"]),
        v2=foo2(udf_alltypes["double_col"], udf_alltypes["int_col"]),
    ).execute()

    expected = udf_df.assign(
        v1=udf_df["double_col"] + udf_df["int_col"],
        v2=udf_df["double_col"] + udf_df["int_col"],
    )

    udf_backend.assert_frame_equal(result, expected)


def test_valid_args_and_kwargs(udf_backend, udf_alltypes, udf_df):
    # Test UDFs with both *args and keyword arguments
    with pytest.warns(FutureWarning, match="v9.0"):

        @elementwise(input_type=[dt.double, dt.int32], output_type=dt.double)
        def foo1(*args, amount):
            # UDF with *args and a keyword-only argument
            return args[0] + args[1] + amount

        @elementwise(input_type=[dt.double, dt.int32], output_type=dt.double)
        def foo2(*args, **kwargs):
            # UDF with *args and **kwargs
            return args[0] + args[1] + kwargs.get("amount", 1)

        @elementwise(input_type=[dt.double, dt.int32], output_type=dt.double)
        def foo3(v, *args, amount):
            # UDF with an explicit positional argument, *args, and a keyword-only
            # argument
            return v + args[0] + amount

        @elementwise(input_type=[dt.double, dt.int32], output_type=dt.double)
        def foo4(v, *args, **kwargs):
            # UDF with an explicit positional argument, *args, and **kwargs
            return v + args[0] + kwargs.get("amount", 1)

    result = udf_alltypes.mutate(
        v1=foo1(udf_alltypes["double_col"], udf_alltypes["int_col"], amount=2),
        v2=foo2(udf_alltypes["double_col"], udf_alltypes["int_col"], amount=2),
        v3=foo3(udf_alltypes["double_col"], udf_alltypes["int_col"], amount=2),
        v4=foo4(udf_alltypes["double_col"], udf_alltypes["int_col"], amount=2),
    ).execute()

    expected = udf_df.assign(
        v1=udf_df["double_col"] + udf_df["int_col"] + 2,
        v2=udf_df["double_col"] + udf_df["int_col"] + 2,
        v3=udf_df["double_col"] + udf_df["int_col"] + 2,
        v4=udf_df["double_col"] + udf_df["int_col"] + 2,
    )

    udf_backend.assert_frame_equal(result, expected)


def test_invalid_kwargs():
    # Test that defining a UDF with a non-column argument that is not a
    # keyword argument raises an error

    with pytest.raises(TypeError, match=".*must be defined as keyword only.*"):
        with pytest.warns(FutureWarning, match="v9.0"):

            @elementwise(input_type=[dt.double], output_type=dt.double)
            def _(v, _):
                return v + 1


@pytest.mark.parametrize("udf", add_one_struct_udfs)
def test_elementwise_udf_destruct(udf_backend, udf_alltypes, udf):
    result = udf_alltypes.mutate(
        udf(udf_alltypes["double_col"]).destructure()
    ).execute()

    expected = udf_alltypes.mutate(
        col1=udf_alltypes["double_col"] + 1,
        col2=udf_alltypes["double_col"] + 2,
    ).execute()

    udf_backend.assert_frame_equal(result, expected)


def test_elementwise_udf_overwrite_destruct(udf_backend, udf_alltypes):
    result = udf_alltypes.mutate(
        overwrite_struct_elementwise(udf_alltypes["double_col"]).destructure()
    ).execute()

    expected = udf_alltypes.mutate(
        double_col=udf_alltypes["double_col"] + 1,
        col2=udf_alltypes["double_col"] + 2,
    ).execute()

    # TODO issue #2649
    # Due to a known limitation with how we treat DestructColumn
    # in assignments, the ordering of op.selections may not exactly
    # correspond with the column ordering we want (i.e. all new columns
    # should appear at the end, but currently they are materialized
    # directly after those overwritten columns).
    udf_backend.assert_frame_equal(result, expected, check_like=True)


def test_elementwise_udf_overwrite_destruct_and_assign(udf_backend, udf_alltypes):
    result = (
        udf_alltypes.mutate(
            overwrite_struct_elementwise(udf_alltypes["double_col"]).destructure()
        )
        .mutate(col3=udf_alltypes["int_col"] * 3)
        .execute()
    )

    expected = udf_alltypes.mutate(
        double_col=udf_alltypes["double_col"] + 1,
        col2=udf_alltypes["double_col"] + 2,
        col3=udf_alltypes["int_col"] * 3,
    ).execute()

    # TODO issue #2649
    # Due to a known limitation with how we treat DestructColumn
    # in assignments, the ordering of op.selections may not exactly
    # correspond with the column ordering we want (i.e. all new columns
    # should appear at the end, but currently they are materialized
    # directly after those overwritten columns).
    udf_backend.assert_frame_equal(result, expected, check_like=True)


@pytest.mark.parametrize("method", ["destructure", "lift", "unpack"])
def test_elementwise_udf_destructure_exact_once(udf_alltypes, method, tmp_path):
    with pytest.warns(FutureWarning, match="v9.0"):

        @elementwise(
            input_type=[dt.double],
            output_type=dt.Struct({"col1": dt.double, "col2": dt.double}),
        )
        def add_one_struct_exact_once(v):
            key = v.iloc[0]
            path = tmp_path / str(key)
            assert not path.exists()
            path.touch()
            return v + 1, v + 2

    struct = add_one_struct_exact_once(udf_alltypes["id"])

    if method == "destructure":
        expr = udf_alltypes.mutate(struct.destructure())
    elif method == "lift":
        expr = udf_alltypes.mutate(struct=struct).struct.lift()
    elif method == "unpack":
        expr = udf_alltypes.mutate(struct=struct).unpack("struct")
    else:
        raise ValueError(f"Invalid method {method}")
    result = expr.execute()

    assert len(result) > 0


def test_elementwise_udf_multiple_overwrite_destruct(udf_backend, udf_alltypes):
    result = udf_alltypes.mutate(
        multiple_overwrite_struct_elementwise(udf_alltypes["double_col"]).destructure()
    ).execute()

    expected = udf_alltypes.mutate(
        double_col=udf_alltypes["double_col"] + 1,
        col2=udf_alltypes["double_col"] + 2,
        float_col=udf_alltypes["double_col"] + 3,
    ).execute()

    # TODO issue #2649
    # Due to a known limitation with how we treat DestructColumn
    # in assignments, the ordering of op.selections may not exactly
    # correspond with the column ordering we want (i.e. all new columns
    # should appear at the end, but currently they are materialized
    # directly after those overwritten columns).
    udf_backend.assert_frame_equal(result, expected, check_like=True)


def test_elementwise_udf_named_destruct(udf_alltypes):
    """Test error when assigning name to a destruct column."""

    add_one_struct_udf = create_add_one_struct_udf(
        result_formatter=lambda v1, v2: (v1, v2)
    )
    with pytest.raises(com.InputTypeError, match="Unable to infer datatype"):
        udf_alltypes.mutate(
            new_struct=add_one_struct_udf(udf_alltypes["double_col"]).destructure()
        )


def test_elementwise_udf_struct(udf_backend, udf_alltypes):
    add_one_struct_udf = create_add_one_struct_udf(
        result_formatter=lambda v1, v2: (v1, v2)
    )
    result = udf_alltypes.mutate(
        new_col=add_one_struct_udf(udf_alltypes["double_col"])
    ).execute()
    result = result.assign(
        col1=result["new_col"].apply(lambda x: x["col1"]),
        col2=result["new_col"].apply(lambda x: x["col2"]),
    )
    result = result.drop("new_col", axis=1)
    expected = udf_alltypes.mutate(
        col1=udf_alltypes["double_col"] + 1,
        col2=udf_alltypes["double_col"] + 2,
    ).execute()

    udf_backend.assert_frame_equal(result, expected)


@pytest.mark.parametrize("udf", demean_struct_udfs)
@pytest.mark.notimpl(["pyspark"])
def test_analytic_udf_destruct(udf_backend, udf_alltypes, udf):
    w = ibis.window(preceding=None, following=None, group_by="year")

    result = udf_alltypes.mutate(
        udf(udf_alltypes["double_col"], udf_alltypes["int_col"]).over(w).destructure()
    ).execute()

    expected = udf_alltypes.mutate(
        demean=udf_alltypes["double_col"] - udf_alltypes["double_col"].mean().over(w),
        demean_weight=udf_alltypes["int_col"] - udf_alltypes["int_col"].mean().over(w),
    ).execute()
    udf_backend.assert_frame_equal(result, expected)


@pytest.mark.notimpl(["pyspark"])
def test_analytic_udf_destruct_no_group_by(udf_backend, udf_alltypes):
    w = ibis.window(preceding=None, following=None)

    demean_struct_udf = create_demean_struct_udf(
        result_formatter=lambda v1, v2: (v1, v2)
    )
    result = udf_alltypes.mutate(
        demean_struct_udf(udf_alltypes["double_col"], udf_alltypes["int_col"])
        .over(w)
        .destructure()
    ).execute()

    expected = udf_alltypes.mutate(
        demean=udf_alltypes["double_col"] - udf_alltypes["double_col"].mean().over(w),
        demean_weight=udf_alltypes["int_col"] - udf_alltypes["int_col"].mean().over(w),
    ).execute()

    udf_backend.assert_frame_equal(result, expected)


@pytest.mark.notimpl(["pyspark"])
def test_analytic_udf_destruct_overwrite(udf_backend, udf_alltypes):
    w = ibis.window(preceding=None, following=None, group_by="year")

    result = udf_alltypes.mutate(
        overwrite_struct_analytic(udf_alltypes["double_col"], udf_alltypes["int_col"])
        .over(w)
        .destructure()
    ).execute()

    expected = udf_alltypes.mutate(
        double_col=udf_alltypes["double_col"]
        - udf_alltypes["double_col"].mean().over(w),
        demean_weight=udf_alltypes["int_col"] - udf_alltypes["int_col"].mean().over(w),
    ).execute()

    # TODO issue #2649
    # Due to a known limitation with how we treat DestructColumn
    # in assignments, the ordering of op.selections may not exactly
    # correspond with the column ordering we want (i.e. all new columns
    # should appear at the end, but currently they are materialized
    # directly after those overwritten columns).
    udf_backend.assert_frame_equal(result, expected, check_like=True)


@pytest.mark.parametrize("udf", mean_struct_udfs)
@pytest.mark.notimpl(["pyspark"])
def test_reduction_udf_destruct_group_by(udf_backend, udf_alltypes, udf):
    result = (
        udf_alltypes.group_by("year")
        .aggregate(
            udf(udf_alltypes["double_col"], udf_alltypes["int_col"]).destructure()
        )
        .execute()
    ).sort_values("year")

    expected = (
        udf_alltypes.group_by("year")
        .aggregate(
            mean=udf_alltypes["double_col"].mean(),
            mean_weight=udf_alltypes["int_col"].mean(),
        )
        .execute()
    ).sort_values("year")

    udf_backend.assert_frame_equal(result, expected)


@pytest.mark.notimpl(["pyspark"])
def test_reduction_udf_destruct_no_group_by(udf_backend, udf_alltypes):
    mean_struct_udf = create_mean_struct_udf(result_formatter=lambda v1, v2: (v1, v2))
    result = udf_alltypes.aggregate(
        mean_struct_udf(
            udf_alltypes["double_col"], udf_alltypes["int_col"]
        ).destructure()
    ).execute()

    expected = udf_alltypes.aggregate(
        mean=udf_alltypes["double_col"].mean(),
        mean_weight=udf_alltypes["int_col"].mean(),
    ).execute()
    udf_backend.assert_frame_equal(result, expected)


@pytest.mark.notimpl(["pyspark"])
def test_reduction_udf_destruct_no_group_by_overwrite(udf_backend, udf_alltypes):
    result = udf_alltypes.aggregate(
        overwrite_struct_reduction(
            udf_alltypes["double_col"], udf_alltypes["int_col"]
        ).destructure()
    ).execute()

    expected = udf_alltypes.aggregate(
        double_col=udf_alltypes["double_col"].mean(),
        mean_weight=udf_alltypes["int_col"].mean(),
    ).execute()
    # TODO issue #2649
    # Due to a known limitation with how we treat DestructColumn
    # in assignments, the ordering of op.selections may not exactly
    # correspond with the column ordering we want (i.e. all new columns
    # should appear at the end, but currently they are materialized
    # directly after those overwritten columns).
    udf_backend.assert_frame_equal(result, expected, check_like=True)


# TODO - windowing - #2553
@pytest.mark.notimpl(["pyspark"])
def test_reduction_udf_destruct_window(udf_backend, udf_alltypes):
    win = ibis.window(
        preceding=ibis.interval(hours=2),
        following=0,
        group_by="year",
        order_by="timestamp_col",
    )
    mean_struct_udf = create_mean_struct_udf(result_formatter=lambda v1, v2: (v1, v2))

    result = udf_alltypes.mutate(
        mean_struct_udf(udf_alltypes["double_col"], udf_alltypes["int_col"])
        .over(win)
        .destructure()
    ).execute()

    expected = udf_alltypes.mutate(
        mean=udf_alltypes["double_col"].mean().over(win),
        mean_weight=udf_alltypes["int_col"].mean().over(win),
    ).execute()

    udf_backend.assert_frame_equal(result, expected)



from __future__ import annotations

import abc
import os
from datetime import datetime, timedelta, timezone
from pathlib import Path
from typing import TYPE_CHECKING, Any
from unittest import mock

import numpy as np
import pandas as pd
import pytest
from filelock import FileLock

import ibis
from ibis import util
from ibis.backends.conftest import TEST_TABLES
from ibis.backends.pyspark import Backend
from ibis.backends.pyspark.datatypes import PySparkSchema
from ibis.backends.tests.base import BackendTest, ServiceBackendTest
from ibis.backends.tests.data import json_types, topk, win
from ibis.conftest import IS_SPARK_REMOTE, SPARK_REMOTE

if TYPE_CHECKING:
    from collections.abc import Iterable


def set_pyspark_database(con, database):
    con._session.catalog.setCurrentDatabase(database)


class BaseSparkTestConf(abc.ABC):
    @property
    @abc.abstractmethod
    def parquet_dir(self) -> str:
        """Directory containing Parquet files."""

    def _load_data(self, **_: Any) -> None:
        import pyspark.sql.functions as F
        import pyspark.sql.types as pt
        from pyspark.sql import Row

        s = self.connection._session
        num_partitions = 4

        sort_cols = {"functional_alltypes": "id"}

        for name in TEST_TABLES:
            path = os.path.join(self.parquet_dir, f"{name}.parquet")
            t = s.read.parquet(path).repartition(num_partitions)
            if (sort_col := sort_cols.get(name)) is not None:
                t = t.sort(sort_col)
            t.createOrReplaceTempView(name)

        s.createDataFrame([(1, "a")], ["foo", "bar"]).createOrReplaceTempView("simple")

        s.createDataFrame(
            [
                Row(abc=Row(a=1.0, b="banana", c=2)),
                Row(abc=Row(a=2.0, b="apple", c=3)),
                Row(abc=Row(a=3.0, b="orange", c=4)),
                Row(abc=Row(a=None, b="banana", c=2)),
                Row(abc=Row(a=2.0, b=None, c=3)),
                Row(abc=None),
                Row(abc=Row(a=3.0, b="orange", c=None)),
            ],
        ).createOrReplaceTempView("struct")

        s.createDataFrame(
            [([1, 2], [[3, 4], [5, 6]], {"a": [[2, 4], [3, 5]]})],
            [
                "list_of_ints",
                "list_of_list_of_ints",
                "map_string_list_of_list_of_ints",
            ],
        ).createOrReplaceTempView("nested_types")
        s.createDataFrame(
            [
                (
                    [1, 2, 3],
                    ["a", "b", "c"],
                    [1.0, 2.0, 3.0],
                    "a",
                    1.0,
                    [[], [1, 2, 3], None],
                ),
                ([4, 5], ["d", "e"], [4.0, 5.0], "a", 2.0, []),
                ([6, None], ["f", None], [6.0, None], "a", 3.0, [None, [], None]),
                (
                    [None, 1, None],
                    [None, "a", None],
                    [],
                    "b",
                    4.0,
                    [[1], [2], [], [3, 4, 5]],
                ),
                ([2, None, 3], ["b", None, "c"], None, "b", 5.0, None),
                (
                    [4, None, None, 5],
                    ["d", None, None, "e"],
                    [4.0, None, None, 5.0],
                    "c",
                    6.0,
                    [[1, 2, 3]],
                ),
            ],
            ["x", "y", "z", "grouper", "scalar_column", "multi_dim"],
        ).createOrReplaceTempView("array_types")

        s.createDataFrame(
            [({(1, 3): [[2, 4], [3, 5]]},)], ["map_tuple_list_of_list_of_ints"]
        ).createOrReplaceTempView("complicated")

        s.createDataFrame(
            [("a", 1, 4.0, "a"), ("b", 2, 5.0, "a"), ("c", 3, 6.0, "b")],
            ["a", "b", "c", "key"],
        ).createOrReplaceTempView("udf")

        s.createDataFrame(
            pd.DataFrame(
                {
                    "a": np.arange(10, dtype=float),
                    "b": [3.0, np.nan] * 5,
                    "key": list("ddeefffggh"),
                }
            )
        ).createOrReplaceTempView("udf_nan")

        s.createDataFrame(
            [(float(i), None if i % 2 else 3.0, "ddeefffggh"[i]) for i in range(10)],
            ["a", "b", "key"],
        ).createOrReplaceTempView("udf_null")

        s.createDataFrame(
            pd.DataFrame(
                {
                    "a": np.arange(4.0).tolist() + np.random.rand(3).tolist(),
                    "b": np.arange(4.0).tolist() + np.random.rand(3).tolist(),
                    "key": list("ddeefff"),
                }
            )
        ).createOrReplaceTempView("udf_random")

        s.createDataFrame(json_types).createOrReplaceTempView("json_t")
        s.createDataFrame(win).createOrReplaceTempView("win")
        s.createDataFrame(topk.to_pandas()).createOrReplaceTempView("topk")

        s.range(0, 10).withColumn("str_col", F.lit("value")).createTempView(
            "basic_table"
        )

        df_nulls = s.createDataFrame(
            [
                ["k1", np.nan, "Alfred", None],
                ["k1", 3.0, None, "joker"],
                ["k2", 27.0, "Batman", "batmobile"],
                ["k2", None, "Catwoman", "motorcycle"],
            ],
            ["key", "age", "user", "toy"],
        )
        df_nulls.createTempView("null_table")

        df_dates = s.createDataFrame(
            [["2018-01-02"], ["2018-01-03"], ["2018-01-04"]], ["date_str"]
        )
        df_dates.createTempView("date_table")

        df_time_indexed = s.createDataFrame(
            [
                [datetime(2017, 1, 2, 5, tzinfo=timezone.utc), 1, 1.0],
                [datetime(2017, 1, 2, 5, tzinfo=timezone.utc), 2, 2.0],
                [datetime(2017, 1, 2, 6, tzinfo=timezone.utc), 1, 3.0],
                [datetime(2017, 1, 2, 6, tzinfo=timezone.utc), 2, 4.0],
                [datetime(2017, 1, 2, 7, tzinfo=timezone.utc), 1, 5.0],
                [datetime(2017, 1, 2, 7, tzinfo=timezone.utc), 2, 6.0],
                [datetime(2017, 1, 4, 8, tzinfo=timezone.utc), 1, 7.0],
                [datetime(2017, 1, 4, 8, tzinfo=timezone.utc), 2, 8.0],
            ],
            ["time", "key", "value"],
        )

        df_time_indexed.createTempView("time_indexed_table")

        if not IS_SPARK_REMOTE:
            # TODO(cpcloud): understand why this doesn't work with spark connect
            df_interval = s.createDataFrame(
                [
                    [
                        timedelta(days=10),
                        timedelta(hours=10),
                        timedelta(minutes=10),
                        timedelta(seconds=10),
                    ]
                ],
                pt.StructType(
                    [
                        pt.StructField(
                            "interval_day",
                            pt.DayTimeIntervalType(
                                pt.DayTimeIntervalType.DAY, pt.DayTimeIntervalType.DAY
                            ),
                        ),
                        pt.StructField(
                            "interval_hour",
                            pt.DayTimeIntervalType(
                                pt.DayTimeIntervalType.HOUR, pt.DayTimeIntervalType.HOUR
                            ),
                        ),
                        pt.StructField(
                            "interval_minute",
                            pt.DayTimeIntervalType(
                                pt.DayTimeIntervalType.MINUTE,
                                pt.DayTimeIntervalType.MINUTE,
                            ),
                        ),
                        pt.StructField(
                            "interval_second",
                            pt.DayTimeIntervalType(
                                pt.DayTimeIntervalType.SECOND,
                                pt.DayTimeIntervalType.SECOND,
                            ),
                        ),
                    ]
                ),
            )

            df_interval.createTempView("interval_table")


if IS_SPARK_REMOTE:

    class TestConf(BaseSparkTestConf, ServiceBackendTest):
        deps = ("pyspark",)
        data_volume = "/data"
        service_name = "spark-connect"

        @property
        def parquet_dir(self) -> str:
            return self.data_volume

        @property
        def test_files(self) -> Iterable[Path]:
            return self.data_dir.joinpath("parquet").glob("*.parquet")

        @staticmethod
        def connect(*, tmpdir, worker_id, **kw):
            from pyspark.sql import SparkSession

            spark = (
                SparkSession.builder.appName("ibis_testing")
                .remote(SPARK_REMOTE)
                .getOrCreate()
            )
            return ibis.pyspark.connect(spark, **kw)

    @pytest.fixture(scope="session")
    def con_streaming(data_dir, tmp_path_factory, worker_id):
        pytest.skip("Streaming tests are not supported in remote mode")

    def write_to_memory(self, expr, table_name):
        assert self.mode == "batch"
        raise NotImplementedError
else:

    class TestConf(BaseSparkTestConf, BackendTest):
        deps = ("pyspark",)

        @property
        def parquet_dir(self) -> str:
            return str(self.data_dir / "parquet")

        @staticmethod
        def connect(*, tmpdir, worker_id, **kw):
            from pyspark.sql import SparkSession

            config = SparkSession.builder.appName("ibis_testing")

            # load from properties file, yuck
            with Path(
                os.environ.get(
                    "SPARK_CONFIG",
                    Path(ibis.__file__)
                    .parents[1]
                    .joinpath("docker", "spark-connect", "conf.properties"),
                )
            ).open(mode="r") as config_file:
                for line in config_file:
                    config = config.config(*map(str.strip, line.strip().split("=", 1)))

            config = (
                config.config("spark.cores.max", "1")
                .config("spark.default.parallelism", "1")
                .config("spark.dynamicAllocation.enabled", "false")
                .config("spark.executor.heartbeatInterval", "3600s")
                .config("spark.executor.instances", "1")
                .config("spark.network.timeout", "4200s")
                .config("spark.rdd.compress", "false")
                .config(
                    "spark.serializer", "org.apache.spark.serializer.KryoSerializer"
                )
                .config("spark.shuffle.compress", "false")
                .config("spark.shuffle.spill.compress", "false")
                .config("spark.sql.execution.arrow.pyspark.enabled", "false")
                .config("spark.sql.shuffle.partitions", "1")
                .config("spark.storage.blockManagerSlaveTimeoutMs", "4200s")
            )

            try:
                from delta.pip_utils import configure_spark_with_delta_pip
            except ImportError:
                configure_spark_with_delta_pip = lambda cfg: cfg
            else:
                config = config.config(
                    "spark.sql.catalog.spark_catalog",
                    "org.apache.spark.sql.delta.catalog.DeltaCatalog",
                ).config(
                    "spark.sql.extensions",
                    "io.delta.sql.DeltaSparkSessionExtension",
                )

            spark = configure_spark_with_delta_pip(config).getOrCreate()
            return ibis.pyspark.connect(spark, **kw)

    class TestConfForStreaming(BackendTest):
        deps = ("pyspark",)

        def _load_data(self, **_: Any) -> None:
            s = self.connection._session
            num_partitions = 4

            watermark_cols = {"functional_alltypes": "timestamp_col"}

            for name, schema in TEST_TABLES.items():
                path = str(self.data_dir / "directory" / "parquet" / name)
                t = (
                    s.readStream.schema(PySparkSchema.from_ibis(schema))
                    .parquet(path)
                    .repartition(num_partitions)
                )
                if (watermark_col := watermark_cols.get(name)) is not None:
                    t = t.withWatermark(watermark_col, "10 seconds")
                t.createOrReplaceTempView(name)

        @classmethod
        def load_data(
            cls, data_dir: Path, tmpdir: Path, worker_id: str, **kw: Any
        ) -> BackendTest:
            """Load testdata from `data_dir`."""
            # handling for multi-processes pytest

            # get the temp directory shared by all workers
            root_tmp_dir = tmpdir.getbasetemp() / "streaming"
            if worker_id != "master":
                root_tmp_dir = root_tmp_dir.parent

            fn = root_tmp_dir / cls.name()
            with FileLock(f"{fn}.lock"):
                cls.skip_if_missing_deps()

                inst = cls(data_dir=data_dir, tmpdir=tmpdir, worker_id=worker_id, **kw)

                if inst.stateful:
                    inst.stateful_load(fn, **kw)
                else:
                    inst.stateless_load(**kw)
                inst.postload(tmpdir=tmpdir, worker_id=worker_id, **kw)
                return inst

        @staticmethod
        def connect(*, tmpdir, worker_id, **kw):
            from pyspark.sql import SparkSession

            # SparkContext is shared globally; only one SparkContext should be active
            # per JVM. We need to create a new SparkSession for streaming tests but
            # this session shares the same SparkContext.
            spark = SparkSession.getActiveSession().newSession()
            con = ibis.pyspark.connect(spark, mode="streaming", **kw)
            return con

    @pytest.fixture(scope="session")
    def con_streaming(data_dir, tmp_path_factory, worker_id):
        backend_test = TestConfForStreaming.load_data(
            data_dir, tmp_path_factory, worker_id
        )
        return backend_test.connection

    @pytest.fixture(autouse=True, scope="function")
    def stop_active_jobs(con_streaming):
        yield
        for sq in con_streaming._session.streams.active:
            sq.stop()
            sq.awaitTermination()

    def write_to_memory(self, expr, table_name):
        assert self.mode == "streaming"
        df = self._session.sql(expr.compile())
        df.writeStream.format("memory").queryName(table_name).start()


@pytest.fixture(scope="session")
def con(data_dir, tmp_path_factory, worker_id):
    backend_test = TestConf.load_data(data_dir, tmp_path_factory, worker_id)
    con = backend_test.connection

    return con


class IbisWindow:
    # Test util class to generate different types of ibis windows
    def __init__(self, windows):
        self.windows = windows

    def get_windows(self):
        # Return a list of Ibis windows
        return [
            ibis.window(preceding=w[0], following=w[1], order_by="time", group_by="key")
            for w in self.windows
        ]


@pytest.fixture
def ibis_windows(request):
    return IbisWindow(request.param).get_windows()


@pytest.fixture(scope="session", autouse=True)
def test_data_db(con):
    name = os.environ.get("IBIS_TEST_DATA_DB", "ibis_testing")
    con.create_database(name, force=True)
    set_pyspark_database(con, name)
    yield name
    con.drop_database(name, force=True)


@pytest.fixture
def temp_database(con, test_data_db):
    name = util.gen_name("database")
    con.create_database(name)
    yield name
    set_pyspark_database(con, test_data_db)
    con.drop_database(name, force=True)


@pytest.fixture(scope="session")
def alltypes(con):
    return con.table("functional_alltypes")


@pytest.fixture
def temp_table_db(con, temp_database):
    name = util.gen_name("table")
    yield temp_database, name
    assert name in con.list_tables(database=temp_database), name
    con.drop_table(name, database=temp_database)


@pytest.fixture(scope="session", autouse=True)
def default_session_fixture():
    with mock.patch.object(Backend, "write_to_memory", write_to_memory, create=True):
        yield


@pytest.fixture
def awards_players_schema():
    return TEST_TABLES["awards_players"]



from __future__ import annotations

import ibis


def test_rank_no_window_frame(con, snapshot):
    t = ibis.table(schema=dict(color=str, price=int), name="diamonds_sample")
    expr = t.mutate(ibis.rank().over(group_by="color", order_by="price"))
    sql = ibis.to_sql(expr, dialect="mssql")

    snapshot.assert_match(sql, "out.sql")



from __future__ import annotations

__all__ = [
    "BigQueryCompiler",
    "ClickHouseCompiler",
    "DataFusionCompiler",
    "DatabricksCompiler",
    "DruidCompiler",
    "DuckDBCompiler",
    "ExasolCompiler",
    "FlinkCompiler",
    "ImpalaCompiler",
    "MSSQLCompiler",
    "MySQLCompiler",
    "OracleCompiler",
    "PostgresCompiler",
    "PySparkCompiler",
    "RisingWaveCompiler",
    "SQLiteCompiler",
    "SnowflakeCompiler",
    "TrinoCompiler",
]

from ibis.backends.sql.compilers.bigquery import BigQueryCompiler
from ibis.backends.sql.compilers.clickhouse import ClickHouseCompiler
from ibis.backends.sql.compilers.databricks import DatabricksCompiler
from ibis.backends.sql.compilers.datafusion import DataFusionCompiler
from ibis.backends.sql.compilers.druid import DruidCompiler
from ibis.backends.sql.compilers.duckdb import DuckDBCompiler
from ibis.backends.sql.compilers.exasol import ExasolCompiler
from ibis.backends.sql.compilers.flink import FlinkCompiler
from ibis.backends.sql.compilers.impala import ImpalaCompiler
from ibis.backends.sql.compilers.mssql import MSSQLCompiler
from ibis.backends.sql.compilers.mysql import MySQLCompiler
from ibis.backends.sql.compilers.oracle import OracleCompiler
from ibis.backends.sql.compilers.postgres import PostgresCompiler
from ibis.backends.sql.compilers.pyspark import PySparkCompiler
from ibis.backends.sql.compilers.risingwave import RisingWaveCompiler
from ibis.backends.sql.compilers.snowflake import SnowflakeCompiler
from ibis.backends.sql.compilers.sqlite import SQLiteCompiler
from ibis.backends.sql.compilers.trino import TrinoCompiler



"""The Microsoft Sql Server backend."""

from __future__ import annotations

import contextlib
import datetime
import struct
from contextlib import closing
from operator import itemgetter
from typing import TYPE_CHECKING, Any
from urllib.parse import unquote_plus

import pyodbc
import sqlglot as sg
import sqlglot.expressions as sge

import ibis
import ibis.backends.sql.compilers as sc
import ibis.common.exceptions as com
import ibis.expr.datatypes as dt
import ibis.expr.operations as ops
import ibis.expr.schema as sch
import ibis.expr.types as ir
from ibis import util
from ibis.backends import CanCreateCatalog, CanCreateDatabase
from ibis.backends.sql import SQLBackend
from ibis.backends.sql.compilers.base import STAR, C

if TYPE_CHECKING:
    from collections.abc import Iterable, Mapping
    from urllib.parse import ParseResult

    import pandas as pd
    import polars as pl
    import pyarrow as pa


def datetimeoffset_to_datetime(value):
    """Convert a datetimeoffset value to a datetime.

    Adapted from https://github.com/mkleehammer/pyodbc/issues/1141
    """
    # ref: https://github.com/mkleehammer/pyodbc/issues/134#issuecomment-281739794
    year, month, day, hour, minute, second, frac, tz_hour, tz_minutes = struct.unpack(
        "<6hI2h", value
    )  # e.g., (2017, 3, 16, 10, 35, 18, 500000000, -6, 0)
    return datetime.datetime(
        year,
        month,
        day,
        hour,
        minute,
        second,
        frac // 1000,
        datetime.timezone(datetime.timedelta(hours=tz_hour, minutes=tz_minutes)),
    )


# For testing we use the collation "Latin1_General_100_BIN2_UTF8"
# which is case-sensitive and supports UTF8.
# This allows us to (hopefully) support both case-sensitive and case-insensitive
# collations.
# It DOES mean, though, that we need to be correct in our usage of case when
# referring to system tables and views.
# So, the correct casing for the tables and views we use often (and the
# corresponding columns):
#
#
# Info schema tables:
# - INFORMATION_SCHEMA.COLUMNS
# - INFORMATION_SCHEMA.SCHEMATA
# - INFORMATION_SCHEMA.TABLES
# Temp table location: tempdb.dbo
# Catalogs: sys.databases
# Databases: sys.schemas


class Backend(SQLBackend, CanCreateCatalog, CanCreateDatabase):
    name = "mssql"
    compiler = sc.mssql.compiler
    supports_create_or_replace = False

    @property
    def version(self) -> str:
        with self._safe_raw_sql("SELECT @@VERSION") as cur:
            [(version,)] = cur.fetchall()
        return version

    def do_connect(
        self,
        host: str = "localhost",
        user: str | None = None,
        password: str | None = None,
        port: int = 1433,
        database: str | None = None,
        driver: str | None = None,
        **kwargs: Any,
    ) -> None:
        """Connect to MSSQL database.

        Parameters
        ----------
        host
            Address of MSSQL server to connect to.
        user
            Username.  Leave blank to use Integrated Authentication.
        password
            Password.  Leave blank to use Integrated Authentication.
        port
            Port of MSSQL server to connect to.
        database
            The MSSQL database to connect to.
        driver
            ODBC Driver to use.

            On Mac and Linux this is usually 'FreeTDS'.

            On Windows, it is usually one of:

            - ODBC Driver 11 for SQL Server
            - ODBC Driver 13 for SQL Server (for both 13 and 13.1)
            - ODBC Driver 17 for SQL Server
            - ODBC Driver 18 for SQL Server

            See https://learn.microsoft.com/en-us/sql/connect/odbc/windows/system-requirements-installation-and-driver-files
        kwargs
            Additional keyword arguments to pass to PyODBC.

        Examples
        --------
        >>> import os
        >>> import ibis
        >>> host = os.environ.get("IBIS_TEST_MSSQL_HOST", "localhost")
        >>> user = os.environ.get("IBIS_TEST_MSSQL_USER", "sa")
        >>> password = os.environ.get("IBIS_TEST_MSSQL_PASSWORD", "1bis_Testing!")
        >>> database = os.environ.get("IBIS_TEST_MSSQL_DATABASE", "ibis_testing")
        >>> driver = os.environ.get("IBIS_TEST_MSSQL_PYODBC_DRIVER", "FreeTDS")
        >>> con = ibis.mssql.connect(
        ...     database=database,
        ...     host=host,
        ...     user=user,
        ...     password=password,
        ...     driver=driver,
        ... )
        >>> con.list_tables()  # doctest: +ELLIPSIS
        [...]
        >>> t = con.table("functional_alltypes")
        >>> t
        DatabaseTable: functional_alltypes
          id              int32
          bool_col        boolean
          tinyint_col     int16
          smallint_col    int16
          int_col         int32
          bigint_col      int64
          float_col       float32
          double_col      float64
          date_string_col string
          string_col      string
          timestamp_col   timestamp(7)
          year            int32
          month           int32
        """

        # If no user/password given, assume Windows Integrated Authentication
        # and set "Trusted_Connection" accordingly
        # see: https://learn.microsoft.com/en-us/sql/connect/odbc/linux-mac/using-integrated-authentication
        if user is None and password is None:
            kwargs.setdefault("Trusted_Connection", "yes")

        if database is not None:
            # passing database=None tries to interpolate "None" into the
            # connection string and use it as a database
            kwargs["database"] = database

        self.con = pyodbc.connect(
            user=user,
            server=f"{host},{port}",
            password=self._escape_special_characters(password),
            driver=driver,
            **kwargs,
        )

        self._post_connect()

    @staticmethod
    def _escape_special_characters(value: str) -> str:
        return "{" + value.replace("}", "}}") + "}"

    @util.experimental
    @classmethod
    def from_connection(cls, con: pyodbc.Connection) -> Backend:
        """Create an Ibis client from an existing connection to a MSSQL database.

        Parameters
        ----------
        con
            An existing connection to a MSSQL database.
        """
        new_backend = cls()
        new_backend._can_reconnect = False
        new_backend.con = con
        new_backend._post_connect()
        return new_backend

    def _post_connect(self):
        # -155 is the code for datetimeoffset
        self.con.add_output_converter(-155, datetimeoffset_to_datetime)

        with closing(self.con.cursor()) as cur:
            cur.execute("SET DATEFIRST 1")

    def _from_url(self, url: ParseResult, **kwargs):
        database, *_ = url.path[1:].split("/", 1)
        kwargs.update(
            {
                "user": url.username,
                "password": unquote_plus(url.password or ""),
                "host": url.hostname,
                "database": database or "",
                "port": url.port or None,
            }
        )

        self._convert_kwargs(kwargs)

        if "host" in kwargs and not kwargs["host"]:
            del kwargs["host"]

        if "user" in kwargs and not kwargs["user"]:
            del kwargs["user"]

        if "password" in kwargs and kwargs["password"] is None:
            del kwargs["password"]

        if "port" in kwargs and kwargs["port"] is None:
            del kwargs["port"]

        if "database" in kwargs and not kwargs["database"]:
            del kwargs["database"]

        if "driver" in kwargs and not kwargs["driver"]:
            del kwargs["driver"]

        return self.connect(**kwargs)

    def get_schema(
        self, name: str, *, catalog: str | None = None, database: str | None = None
    ) -> sch.Schema:
        # TODO: this is brittle and should be improved. We want to be able to
        # identify if a given table is a temp table and update the search
        # location accordingly.
        if name.startswith("ibis_cache_"):
            catalog, database = ("tempdb", "dbo")
            name = "##" + name

        query = (
            sg.select(
                C.column_name,
                C.data_type,
                C.is_nullable,
                C.numeric_precision,
                C.numeric_scale,
                C.datetime_precision,
            )
            .from_(
                sg.table(
                    "COLUMNS",
                    db="INFORMATION_SCHEMA",
                    catalog=catalog or self.current_catalog,
                )
            )
            .where(
                C.table_name.eq(sge.convert(name)),
                C.table_schema.eq(sge.convert(database or self.current_database)),
            )
            .order_by(C.ordinal_position)
        )

        with self._safe_raw_sql(query) as cur:
            meta = cur.fetchall()

        if not meta:
            fqn = sg.table(name, db=database, catalog=catalog).sql(self.dialect)
            raise com.TableNotFound(fqn)

        mapping = {}
        for (
            col,
            typ,
            is_nullable,
            numeric_precision,
            numeric_scale,
            datetime_precision,
        ) in meta:
            newtyp = self.compiler.type_mapper.from_string(
                typ, nullable=is_nullable == "YES"
            )

            if typ == "float":
                newcls = dt.Float64 if numeric_precision == 53 else dt.Float32
                newtyp = newcls(nullable=newtyp.nullable)
            elif newtyp.is_decimal():
                newtyp = newtyp.copy(precision=numeric_precision, scale=numeric_scale)
            elif newtyp.is_timestamp():
                newtyp = newtyp.copy(scale=datetime_precision)
            mapping[col] = newtyp

        return sch.Schema(mapping)

    def _get_schema_using_query(self, query: str) -> sch.Schema:
        # Docs describing usage of dm_exec_describe_first_result_set
        # https://learn.microsoft.com/en-us/sql/relational-databases/system-dynamic-management-views/sys-dm-exec-describe-first-result-set-transact-sql?view=sql-server-ver16
        tsql = sge.convert(str(query)).sql(self.dialect)

        # For some reason when using "Latin1_General_100_BIN2_UTF8"
        # the stored procedure `sp_describe_first_result_set` starts throwing errors about DLL loading.
        # This "dynamic management function" uses the same algorithm and allows
        # us to pre-filter the columns we want back.
        # The syntax is:
        # `sys.dm_exec_describe_first_result_set(@tsql, @params, @include_browse_information)`
        query = f"""
        SELECT
          name,
          is_nullable,
          system_type_name,
          precision,
          scale,
          error_number,
          error_message
        FROM sys.dm_exec_describe_first_result_set(N{tsql}, NULL, 0)
        ORDER BY column_ordinal
        """
        with self._safe_raw_sql(query) as cur:
            rows = cur.fetchall()

        schema = {}
        for (
            name,
            nullable,
            system_type_name,
            precision,
            scale,
            err_num,
            err_msg,
        ) in rows:
            if err_num is not None:
                raise com.IbisInputError(f".sql failed with message: {err_msg}")

            newtyp = self.compiler.type_mapper.from_string(
                system_type_name, nullable=nullable
            )

            if system_type_name == "float":
                newcls = dt.Float64 if precision == 53 else dt.Float32
                newtyp = newcls(nullable=newtyp.nullable)
            elif newtyp.is_decimal():
                newtyp = newtyp.copy(precision=precision, scale=scale)
            elif newtyp.is_timestamp():
                newtyp = newtyp.copy(scale=scale)

            if name is None:
                name = util.gen_name("col")

            schema[name] = newtyp

        return sch.Schema(schema)

    @property
    def current_catalog(self) -> str:
        with self._safe_raw_sql(sg.select(self.compiler.f.db_name())) as cur:
            [(database,)] = cur.fetchall()
        return database

    def list_catalogs(self, like: str | None = None) -> list[str]:
        s = sg.table("databases", db="sys")

        with self._safe_raw_sql(sg.select(C.name).from_(s)) as cur:
            results = list(map(itemgetter(0), cur.fetchall()))

        return self._filter_with_like(results, like=like)

    @property
    def current_database(self) -> str:
        with self._safe_raw_sql(sg.select(self.compiler.f.schema_name())) as cur:
            [(schema,)] = cur.fetchall()
        return schema

    @contextlib.contextmanager
    def begin(self):
        with contextlib.closing(self.con.cursor()) as cur:
            yield cur

    @contextlib.contextmanager
    def _ddl_begin(self):
        con = self.con
        cur = con.cursor()
        try:
            yield cur
        except Exception:
            con.rollback()
            raise
        else:
            con.commit()
        finally:
            cur.close()

    @contextlib.contextmanager
    def _safe_raw_sql(self, query, *args, **kwargs):
        with contextlib.suppress(AttributeError):
            query = query.sql(self.dialect)

        with self.begin() as cur:
            cur.execute(query, *args, **kwargs)
            yield cur

    @contextlib.contextmanager
    def _safe_ddl(self, query, *args, **kwargs):
        with contextlib.suppress(AttributeError):
            query = query.sql(self.dialect)

        with self._ddl_begin() as cur:
            cur.execute(query, *args, **kwargs)
            yield cur

    def raw_sql(self, query: str | sg.Expression, **kwargs: Any) -> Any:
        with contextlib.suppress(AttributeError):
            query = query.sql(self.dialect)

        con = self.con
        cursor = con.cursor()

        cursor.execute(query, **kwargs)
        return cursor

    def create_catalog(self, name: str, force: bool = False) -> None:
        expr = (
            sg.select(STAR)
            .from_(sg.table("databases", db="sys"))
            .where(C.name.eq(sge.convert(name)))
        )
        stmt = sge.Create(
            kind="DATABASE", this=sg.to_identifier(name, quoted=self.compiler.quoted)
        ).sql(self.dialect)
        create_stmt = (
            f"""\
IF NOT EXISTS ({expr.sql(self.dialect)})
BEGIN
  {stmt};
END;
GO"""
            if force
            else stmt
        )
        with self._safe_ddl(create_stmt):
            pass

    def drop_catalog(self, name: str, force: bool = False) -> None:
        with self._safe_ddl(
            sge.Drop(
                kind="DATABASE",
                this=sg.to_identifier(name, quoted=self.compiler.quoted),
                exists=force,
            )
        ):
            pass

    def create_database(
        self, name: str, catalog: str | None = None, force: bool = False
    ) -> None:
        current_catalog = self.current_catalog
        should_switch_catalog = catalog is not None and catalog != current_catalog
        quoted = self.compiler.quoted

        expr = (
            sg.select(STAR)
            .from_(sg.table("schemas", db="sys"))
            .where(C.name.eq(sge.convert(name)))
        )
        stmt = sge.Create(
            kind="SCHEMA", this=sg.to_identifier(name, quoted=quoted)
        ).sql(self.dialect)

        create_stmt = (
            f"""\
IF NOT EXISTS ({expr.sql(self.dialect)})
BEGIN
  {stmt};
END;
GO"""
            if force
            else stmt
        )

        with self.begin() as cur:
            if should_switch_catalog:
                cur.execute(
                    sge.Use(this=sg.to_identifier(catalog, quoted=quoted)).sql(
                        self.dialect
                    )
                )

            cur.execute(create_stmt)

            if should_switch_catalog:
                cur.execute(
                    sge.Use(this=sg.to_identifier(current_catalog, quoted=quoted)).sql(
                        self.dialect
                    )
                )

    def drop_database(
        self, name: str, catalog: str | None = None, force: bool = False
    ) -> None:
        current_catalog = self.current_catalog
        should_switch_catalog = catalog is not None and catalog != current_catalog

        quoted = self.compiler.quoted

        with self.begin() as cur:
            if should_switch_catalog:
                cur.execute(
                    sge.Use(this=sg.to_identifier(catalog, quoted=quoted)).sql(
                        self.dialect
                    )
                )

            cur.execute(
                sge.Drop(
                    kind="SCHEMA",
                    exists=force,
                    this=sg.to_identifier(name, quoted=quoted),
                ).sql(self.dialect)
            )

            if should_switch_catalog:
                cur.execute(
                    sge.Use(this=sg.to_identifier(current_catalog, quoted=quoted)).sql(
                        self.dialect
                    )
                )

    def list_tables(
        self,
        like: str | None = None,
        database: tuple[str, str] | str | None = None,
    ) -> list[str]:
        """List the tables in the database.

        ::: {.callout-note}
        ## Ibis does not use the word `schema` to refer to database hierarchy.

        A collection of tables is referred to as a `database`.
        A collection of `database` is referred to as a `catalog`.

        These terms are mapped onto the corresponding features in each
        backend (where available), regardless of whether the backend itself
        uses the same terminology.
        :::

        Parameters
        ----------
        like
            A pattern to use for listing tables.
        database
            Table location. If not passed, uses the current catalog and database.

            To specify a table in a separate catalog, you can pass in the
            catalog and database as a string `"catalog.database"`, or as a tuple of
            strings `("catalog", "database")`.

        """
        table_loc = self._to_sqlglot_table(database)
        catalog, db = self._to_catalog_db_tuple(table_loc)

        sql = (
            sg.select(C.table_name)
            .from_(
                sg.table(
                    "TABLES",
                    db="INFORMATION_SCHEMA",
                    catalog=catalog if catalog is not None else self.current_catalog,
                )
            )
            .where(C.table_schema.eq(sge.convert(db or self.current_database)))
            .distinct()
        )

        sql = sql.sql(self.dialect)

        with self._safe_raw_sql(sql) as cur:
            out = cur.fetchall()

        return self._filter_with_like(map(itemgetter(0), out), like)

    def list_databases(
        self, like: str | None = None, catalog: str | None = None
    ) -> list[str]:
        query = sg.select(C.schema_name).from_(
            sg.table(
                "SCHEMATA",
                db="INFORMATION_SCHEMA",
                catalog=catalog or self.current_catalog,
            )
        )
        with self._safe_raw_sql(query) as cur:
            results = list(map(itemgetter(0), cur.fetchall()))
        return self._filter_with_like(results, like=like)

    def create_table(
        self,
        name: str,
        obj: ir.Table
        | pd.DataFrame
        | pa.Table
        | pl.DataFrame
        | pl.LazyFrame
        | None = None,
        *,
        schema: sch.SchemaLike | None = None,
        database: str | None = None,
        temp: bool | None = None,
        overwrite: bool = False,
    ) -> ir.Table:
        """Create a new table.

        Parameters
        ----------
        name
            Name of the new table.
        obj
            An Ibis table expression or pandas table that will be used to
            extract the schema and the data of the new table. If not provided,
            `schema` must be given.
        schema
            The schema for the new table. Only one of `schema` or `obj` can be
            provided.
        database
            Name of the database where the table will be created, if not the
            default.

            To specify a location in a separate catalog, you can pass in the
            catalog and database as a string `"catalog.database"`, or as a tuple of
            strings `("catalog", "database")`.
        temp
            Whether a table is temporary or not.
            All created temp tables are "Global Temporary Tables". They will be
            created in "tempdb.dbo" and will be prefixed with "##".
        overwrite
            Whether to clobber existing data.
            `overwrite` and `temp` cannot be used together with MSSQL.

        Returns
        -------
        Table
            The table that was created.

        """
        if obj is None and schema is None:
            raise ValueError("Either `obj` or `schema` must be specified")
        if schema is not None:
            schema = ibis.schema(schema)

        if temp and overwrite:
            raise ValueError(
                "MSSQL doesn't support overwriting temp tables, create a new temp table instead."
            )

        table_loc = self._to_sqlglot_table(database)
        catalog, db = self._to_catalog_db_tuple(table_loc)

        properties = []

        if temp:
            properties.append(sge.TemporaryProperty())
            catalog, db = None, None

        if obj is not None:
            if not isinstance(obj, ir.Expr):
                table = ibis.memtable(obj)
            else:
                table = obj

            self._run_pre_execute_hooks(table)

            query = self.compiler.to_sqlglot(table)
        else:
            query = None

        if overwrite:
            temp_name = util.gen_name(f"{self.name}_table")
        else:
            temp_name = name

        if not schema:
            schema = table.schema()

        quoted = self.compiler.quoted
        raw_table = sg.table(temp_name, catalog=catalog, db=db, quoted=False)
        target = sge.Schema(
            this=sg.table(
                "#" * bool(temp) + temp_name, catalog=catalog, db=db, quoted=quoted
            ),
            expressions=schema.to_sqlglot(self.dialect),
        )

        create_stmt = sge.Create(
            kind="TABLE",
            this=target,
            properties=sge.Properties(expressions=properties),
        )

        this = sg.table(name, catalog=catalog, db=db, quoted=quoted)
        raw_this = sg.table(name, catalog=catalog, db=db, quoted=False)
        with self._safe_ddl(create_stmt) as cur:
            if query is not None:
                # You can specify that a table is temporary for the sqlglot `Create` but not
                # for the subsequent `Insert`, so we need to shove a `#` in
                # front of the table identifier.
                _table = sg.table(
                    "##" * bool(temp) + temp_name,
                    catalog=catalog,
                    db=db,
                    quoted=self.compiler.quoted,
                )
                insert_stmt = sge.Insert(this=_table, expression=query).sql(
                    self.dialect
                )
                cur.execute(insert_stmt)

            if overwrite:
                cur.execute(
                    sge.Drop(kind="TABLE", this=this, exists=True).sql(self.dialect)
                )
                old = raw_table.sql(self.dialect)
                new = raw_this.sql(self.dialect)
                cur.execute(f"EXEC sp_rename '{old}', '{new}'")

        if temp:
            # If a temporary table, amend the output name/catalog/db accordingly
            name = "##" + name
            catalog = "tempdb"
            db = "dbo"

        if schema is None:
            return self.table(name, database=(catalog, db))

        # preserve the input schema if it was provided
        return ops.DatabaseTable(
            name,
            schema=schema,
            source=self,
            namespace=ops.Namespace(catalog=catalog, database=db),
        ).to_expr()

    def _register_in_memory_table(self, op: ops.InMemoryTable) -> None:
        schema = op.schema
        if null_columns := [col for col, dtype in schema.items() if dtype.is_null()]:
            raise com.IbisTypeError(
                "MS SQL cannot yet reliably handle `null` typed columns; "
                f"got null typed columns: {null_columns}"
            )

        name = op.name
        quoted = self.compiler.quoted

        create_stmt = sg.exp.Create(
            kind="TABLE",
            this=sg.exp.Schema(
                this=sg.to_identifier(name, quoted=quoted),
                expressions=schema.to_sqlglot(self.dialect),
            ),
        )

        df = op.data.to_frame()
        data = df.itertuples(index=False)

        insert_stmt = self._build_insert_template(name, schema=schema, columns=True)
        with self._safe_ddl(create_stmt) as cur:
            if not df.empty:
                cur.executemany(insert_stmt, data)

    def _cursor_batches(
        self,
        expr: ir.Expr,
        params: Mapping[ir.Scalar, Any] | None = None,
        limit: int | str | None = None,
        chunk_size: int = 1 << 20,
    ) -> Iterable[list[tuple]]:
        def process_value(value, dtype):
            return bool(value) if dtype.is_boolean() else value

        types = expr.as_table().schema().types

        for batch in super()._cursor_batches(
            expr, params=params, limit=limit, chunk_size=chunk_size
        ):
            yield [tuple(map(process_value, row, types)) for row in batch]



from __future__ import annotations

import concurrent.futures
import os
import tempfile
from contextlib import closing
from pathlib import Path
from typing import TYPE_CHECKING, Any
from urllib.parse import parse_qs, urlparse
from urllib.request import urlretrieve

import pyarrow.parquet as pq
import pyarrow_hotfix  # noqa: F401
import pytest
import snowflake.connector as sc
import sqlglot as sg

import ibis
from ibis.backends.conftest import TEST_TABLES
from ibis.backends.sql.datatypes import SnowflakeType
from ibis.backends.tests.base import BackendTest
from ibis.formats.pyarrow import PyArrowSchema

if TYPE_CHECKING:
    from ibis.backends import BaseBackend


def _get_url():
    if (url := os.environ.get("SNOWFLAKE_URL")) is not None:
        return url
    elif os.environ.get("SNOWFLAKE_HOME"):
        import tomli

        # requires a connection named ibis_testing or one explicitly set with
        # export SNOWFLAKE_DEFAULT_CONNECTION_NAME
        connection_name = os.environ.get(
            "SNOWFLAKE_DEFAULT_CONNECTION_NAME", "ibis_testing"
        )
        config_file = Path(os.environ["SNOWFLAKE_HOME"], "connections.toml")
        config_text = config_file.read_text()
        params = tomli.loads(config_text)[connection_name]
        user, password, account, database, schema, warehouse = (
            params["user"],
            params["password"],
            params["account"],
            params["database"],
            params["schema"],
            params["warehouse"],
        )
        return f"snowflake://{user}:{password}@{account}/{database}/{schema}?warehouse={warehouse}"
    else:
        try:
            user, password, account, database, schema, warehouse = tuple(
                os.environ[f"SNOWFLAKE_{part}"]
                for part in [
                    "USER",
                    "PASSWORD",
                    "ACCOUNT",
                    "DATABASE",
                    "SCHEMA",
                    "WAREHOUSE",
                ]
            )
        except KeyError as e:
            pytest.skip(
                f"missing URL part: {e} or SNOWFLAKE_URL environment variable is not defined"
            )
        return f"snowflake://{user}:{password}@{account}/{database}/{schema}?warehouse={warehouse}"


def copy_into(con, data_dir: Path, table: str) -> None:
    stage = "ibis_testing"
    file = data_dir.joinpath("parquet", f"{table}.parquet").absolute()
    schema = PyArrowSchema.to_ibis(pq.read_metadata(file).schema.to_arrow_schema())
    columns = ", ".join(
        f"$1:{name}{'::VARCHAR' * typ.is_timestamp()}::{SnowflakeType.to_string(typ)}"
        for name, typ in schema.items()
    )
    con.execute(f"PUT {file.as_uri()} @{stage}/{file.name}")
    con.execute(
        f"""
        COPY INTO "{table}"
        FROM (SELECT {columns} FROM @{stage}/{file.name})
        FILE_FORMAT = (TYPE = PARQUET)
        """
    )


class TestConf(BackendTest):
    supports_map = True
    deps = ("snowflake.connector",)
    supports_tpch = True
    supports_tpcds = True

    def _load_tpc(self, *, suite, scale_factor) -> None:
        """Create views of data in the TPC-H catalog that ships with Trino.

        This method create relations that have column names prefixed with the
        first one (or two in the case of partsupp -> ps) character table name
        to match the DuckDB TPC-H query conventions.
        """
        con = self.connection
        schema = f"tpc{suite}"

        con.create_database(schema, force=True)

        parquet_dir = self.data_dir.joinpath(schema, f"sf={scale_factor}", "parquet")
        assert parquet_dir.exists(), parquet_dir

        tables = frozenset(con.list_tables(database=("IBIS_TESTING", schema)))
        for path in parquet_dir.glob("*.parquet"):
            table_name = path.with_suffix("").name
            if table_name not in tables:
                con.create_table(
                    table_name,
                    con.read_parquet(path),
                    database=f'IBIS_TESTING."{schema}"',
                )

    def _transform_tpc_sql(self, parsed, *, suite, leaves):
        def quote(node):
            if isinstance(node, sg.exp.Identifier):
                return sg.to_identifier(node.name, quoted=True)
            return node

        def add_catalog_and_schema(node):
            if isinstance(node, sg.exp.Table) and node.name in leaves:
                node.args["db"] = sg.to_identifier(f"tpc{suite}")
                node.args["catalog"] = sg.to_identifier("IBIS_TESTING")
            return node

        result = parsed.transform(add_catalog_and_schema)
        result = result.transform(quote)
        return result

    def _load_data(self, **_: Any) -> None:
        """Load test data into a Snowflake backend instance."""

        connect_args = {}

        url = urlparse(_get_url())

        if url.path:
            db, schema = url.path[1:].split("/", 1)
            (warehouse,) = parse_qs(url.query)["warehouse"]
            connect_args.update(
                {
                    "user": url.username,
                    "password": url.password,
                    "account": url.hostname,
                    "warehouse": warehouse,
                }
            )
        else:
            db = os.environ["SNOWFLAKE_DATABASE"]
            schema = os.environ["SNOWFLAKE_SCHEMA"]

        dbschema = f"{db}.{schema}"

        with closing(sc.connect(**connect_args)) as con, closing(con.cursor()) as c:
            c.execute("ALTER SESSION SET MULTI_STATEMENT_COUNT = 0 JSON_INDENT = 0")
            c.execute(
                f"""
                CREATE DATABASE IF NOT EXISTS {db};
                CREATE SCHEMA IF NOT EXISTS {dbschema};
                USE {dbschema};
                CREATE TEMP STAGE {db};
                CREATE STAGE IF NOT EXISTS ibis_testing;
                CREATE STAGE IF NOT EXISTS models;
                {self.script_dir.joinpath("snowflake.sql").read_text()}
                """
            )

            with tempfile.TemporaryDirectory() as d:
                path, _ = urlretrieve(
                    "https://storage.googleapis.com/ibis-testing-data/model.joblib",
                    os.path.join(d, "model.joblib"),
                )

                assert os.path.exists(path)
                assert os.path.getsize(path)

                c.execute(f"PUT {Path(path).as_uri()} @MODELS")

            # not much we can do to make this faster, but running these in
            # multiple threads seems to save about 2x
            with concurrent.futures.ThreadPoolExecutor() as exe:
                for future in concurrent.futures.as_completed(
                    exe.submit(copy_into, c, self.data_dir, table)
                    for table in TEST_TABLES.keys()
                ):
                    future.result()

    @staticmethod
    def connect(*, tmpdir, worker_id, **kw) -> BaseBackend:
        if os.environ.get("SNOWFLAKE_SNOWPARK"):
            import snowflake.snowpark as sp

            if connection_name := os.environ.get("SNOWFLAKE_DEFAULT_CONNECTION_NAME"):
                builder = sp.Session.builder.config("connection_name", connection_name)
            else:
                builder = sp.Session.builder.configs(
                    {
                        "user": os.environ["SNOWFLAKE_USER"],
                        "account": os.environ["SNOWFLAKE_ACCOUNT"],
                        "password": os.environ["SNOWFLAKE_PASSWORD"],
                        "warehouse": os.environ["SNOWFLAKE_WAREHOUSE"],
                        "database": os.environ["SNOWFLAKE_DATABASE"],
                        "schema": os.environ["SNOWFLAKE_SCHEMA"],
                    }
                )
            return ibis.backends.snowflake.Backend.from_connection(builder.create())
        else:
            return ibis.connect(_get_url(), **kw)


@pytest.fixture(scope="session")
def con(data_dir, tmp_path_factory, worker_id):
    return TestConf.load_data(data_dir, tmp_path_factory, worker_id).connection



"""PostgreSQL backend."""

from __future__ import annotations

import contextlib
import inspect
from operator import itemgetter
from typing import TYPE_CHECKING, Any
from urllib.parse import unquote_plus

import sqlglot as sg
import sqlglot.expressions as sge
from pandas.api.types import is_float_dtype

import ibis
import ibis.backends.sql.compilers as sc
import ibis.common.exceptions as com
import ibis.common.exceptions as exc
import ibis.expr.datatypes as dt
import ibis.expr.operations as ops
import ibis.expr.schema as sch
import ibis.expr.types as ir
from ibis import util
from ibis.backends import CanCreateDatabase, CanListCatalog
from ibis.backends.sql import SQLBackend
from ibis.backends.sql.compilers.base import TRUE, C, ColGen

if TYPE_CHECKING:
    from collections.abc import Callable
    from urllib.parse import ParseResult

    import pandas as pd
    import polars as pl
    import psycopg2
    import pyarrow as pa


class Backend(SQLBackend, CanListCatalog, CanCreateDatabase):
    name = "postgres"
    compiler = sc.postgres.compiler
    supports_python_udfs = True

    def _from_url(self, url: ParseResult, **kwargs):
        """Connect to a backend using a URL `url`.

        Parameters
        ----------
        url
            URL with which to connect to a backend.
        kwargs
            Additional keyword arguments

        Returns
        -------
        BaseBackend
            A backend instance

        """
        database, *schema = url.path[1:].split("/", 1)
        connect_args = {
            "user": url.username,
            "password": unquote_plus(url.password or ""),
            "host": url.hostname,
            "database": database or "",
            "schema": schema[0] if schema else "",
            "port": url.port,
        }

        kwargs.update(connect_args)
        self._convert_kwargs(kwargs)

        if "user" in kwargs and not kwargs["user"]:
            del kwargs["user"]

        if "host" in kwargs and not kwargs["host"]:
            del kwargs["host"]

        if "database" in kwargs and not kwargs["database"]:
            del kwargs["database"]

        if "schema" in kwargs and not kwargs["schema"]:
            del kwargs["schema"]

        if "password" in kwargs and kwargs["password"] is None:
            del kwargs["password"]

        if "port" in kwargs and kwargs["port"] is None:
            del kwargs["port"]

        return self.connect(**kwargs)

    def _register_in_memory_table(self, op: ops.InMemoryTable) -> None:
        from psycopg2.extras import execute_batch

        schema = op.schema
        if null_columns := [col for col, dtype in schema.items() if dtype.is_null()]:
            raise exc.IbisTypeError(
                f"{self.name} cannot yet reliably handle `null` typed columns; "
                f"got null typed columns: {null_columns}"
            )

        name = op.name
        quoted = self.compiler.quoted
        create_stmt = sg.exp.Create(
            kind="TABLE",
            this=sg.exp.Schema(
                this=sg.to_identifier(name, quoted=quoted),
                expressions=schema.to_sqlglot(self.dialect),
            ),
            properties=sg.exp.Properties(expressions=[sge.TemporaryProperty()]),
        )
        create_stmt_sql = create_stmt.sql(self.dialect)

        df = op.data.to_frame()
        # nan gets compiled into 'NaN'::float which throws errors in non-float columns
        # In order to hold NaN values, pandas automatically converts integer columns
        # to float columns if there are NaN values in them. Therefore, we need to convert
        # them to their original dtypes (that support pd.NA) to figure out which columns
        # are actually non-float, then fill the NaN values in those columns with None.
        convert_df = df.convert_dtypes()
        for col in convert_df.columns:
            if not is_float_dtype(convert_df[col]):
                df[col] = df[col].replace(float("nan"), None)

        data = df.itertuples(index=False)
        sql = self._build_insert_template(
            name, schema=schema, columns=True, placeholder="%s"
        )

        with self.begin() as cur:
            cur.execute(create_stmt_sql)
            execute_batch(cur, sql, data, 128)

    @contextlib.contextmanager
    def begin(self):
        con = self.con
        cursor = con.cursor()
        try:
            yield cursor
        except Exception:
            con.rollback()
            raise
        else:
            con.commit()
        finally:
            cursor.close()

    def _fetch_from_cursor(self, cursor, schema: sch.Schema) -> pd.DataFrame:
        import pandas as pd

        from ibis.backends.postgres.converter import PostgresPandasData

        try:
            df = pd.DataFrame.from_records(
                cursor, columns=schema.names, coerce_float=True
            )
        except Exception:
            # clean up the cursor if we fail to create the DataFrame
            #
            # in the sqlite case failing to close the cursor results in
            # artificially locked tables
            cursor.close()
            raise
        df = PostgresPandasData.convert_table(df, schema)
        return df

    @property
    def version(self):
        version = f"{self.con.server_version:0>6}"
        major = int(version[:2])
        minor = int(version[2:4])
        patch = int(version[4:])
        pieces = [major]
        if minor:
            pieces.append(minor)
        pieces.append(patch)
        return ".".join(map(str, pieces))

    def do_connect(
        self,
        host: str | None = None,
        user: str | None = None,
        password: str | None = None,
        port: int = 5432,
        database: str | None = None,
        schema: str | None = None,
        **kwargs: Any,
    ) -> None:
        """Create an Ibis client connected to PostgreSQL database.

        Parameters
        ----------
        host
            Hostname
        user
            Username
        password
            Password
        port
            Port number
        database
            Database to connect to
        schema
            PostgreSQL schema to use. If `None`, use the default `search_path`.
        kwargs
            Additional keyword arguments to pass to the backend client connection.

        Examples
        --------
        >>> import os
        >>> import ibis
        >>> host = os.environ.get("IBIS_TEST_POSTGRES_HOST", "localhost")
        >>> user = os.environ.get("IBIS_TEST_POSTGRES_USER", "postgres")
        >>> password = os.environ.get("IBIS_TEST_POSTGRES_PASSWORD", "postgres")
        >>> database = os.environ.get("IBIS_TEST_POSTGRES_DATABASE", "ibis_testing")
        >>> con = ibis.postgres.connect(database=database, host=host, user=user, password=password)
        >>> con.list_tables()  # doctest: +ELLIPSIS
        [...]
        >>> t = con.table("functional_alltypes")
        >>> t
        DatabaseTable: functional_alltypes
          id              int32
          bool_col        boolean
          tinyint_col     int16
          smallint_col    int16
          int_col         int32
          bigint_col      int64
          float_col       float32
          double_col      float64
          date_string_col string
          string_col      string
          timestamp_col   timestamp(6)
          year            int32
          month           int32
        """
        import psycopg2
        import psycopg2.extras

        psycopg2.extras.register_default_json(loads=lambda x: x)

        self.con = psycopg2.connect(
            host=host,
            port=port,
            user=user,
            password=password,
            database=database,
            options=(f"-csearch_path={schema}" * (schema is not None)) or None,
            **kwargs,
        )

        self._post_connect()

    @util.experimental
    @classmethod
    def from_connection(cls, con: psycopg2.extensions.connection) -> Backend:
        """Create an Ibis client from an existing connection to a PostgreSQL database.

        Parameters
        ----------
        con
            An existing connection to a PostgreSQL database.
        """
        new_backend = cls()
        new_backend._can_reconnect = False
        new_backend.con = con
        new_backend._post_connect()
        return new_backend

    def _post_connect(self) -> None:
        with self.begin() as cur:
            cur.execute("SET TIMEZONE = UTC")

    @property
    def _session_temp_db(self) -> str | None:
        # Postgres doesn't assign the temporary table database until the first
        # temp table is created in a given session.
        # Before that temp table is created, this will return `None`
        # After a temp table is created, it will return `pg_temp_N` where N is
        # some integer
        res = self.raw_sql(
            "select nspname from pg_namespace where oid = pg_my_temp_schema()"
        ).fetchone()
        if res is not None:
            return res[0]
        return res

    def list_tables(
        self,
        like: str | None = None,
        database: tuple[str, str] | str | None = None,
    ) -> list[str]:
        """List the tables in the database.

        ::: {.callout-note}
        ## Ibis does not use the word `schema` to refer to database hierarchy.

        A collection of tables is referred to as a `database`.
        A collection of `database` is referred to as a `catalog`.

        These terms are mapped onto the corresponding features in each
        backend (where available), regardless of whether the backend itself
        uses the same terminology.
        :::

        Parameters
        ----------
        like
            A pattern to use for listing tables.
        database
            Database to list tables from. Default behavior is to show tables in
            the current database.
        """

        if database is not None:
            table_loc = database
        else:
            table_loc = (self.current_catalog, self.current_database)

        table_loc = self._to_sqlglot_table(table_loc)

        conditions = [TRUE]

        if (db := table_loc.args["db"]) is not None:
            db.args["quoted"] = False
            db = db.sql(dialect=self.name)
            conditions.append(C.table_schema.eq(sge.convert(db)))
        if (catalog := table_loc.args["catalog"]) is not None:
            catalog.args["quoted"] = False
            catalog = catalog.sql(dialect=self.name)
            conditions.append(C.table_catalog.eq(sge.convert(catalog)))

        sql = (
            sg.select("table_name")
            .from_(sg.table("tables", db="information_schema"))
            .distinct()
            .where(*conditions)
            .sql(self.dialect)
        )

        with self._safe_raw_sql(sql) as cur:
            out = cur.fetchall()

        # Include temporary tables only if no database has been explicitly specified
        # to avoid temp tables showing up in all calls to `list_tables`
        if db == "public":
            out += self._fetch_temp_tables()

        return self._filter_with_like(map(itemgetter(0), out), like)

    def _fetch_temp_tables(self):
        # postgres temporary tables are stored in a separate schema
        # so we need to independently grab them and return them along with
        # the existing results

        sql = (
            sg.select("table_name")
            .from_(sg.table("tables", db="information_schema"))
            .distinct()
            .where(C.table_type.eq(sge.convert("LOCAL TEMPORARY")))
            .sql(self.dialect)
        )

        with self._safe_raw_sql(sql) as cur:
            out = cur.fetchall()

        return out

    def list_catalogs(self, like=None) -> list[str]:
        # http://dba.stackexchange.com/a/1304/58517
        cats = (
            sg.select(C.datname)
            .from_(sg.table("pg_database", db="pg_catalog"))
            .where(sg.not_(C.datistemplate))
        )
        with self._safe_raw_sql(cats) as cur:
            catalogs = list(map(itemgetter(0), cur))

        return self._filter_with_like(catalogs, like)

    def list_databases(
        self, *, like: str | None = None, catalog: str | None = None
    ) -> list[str]:
        dbs = sg.select(C.schema_name).from_(
            sg.table("schemata", db="information_schema")
        )
        with self._safe_raw_sql(dbs) as cur:
            databases = list(map(itemgetter(0), cur))

        return self._filter_with_like(databases, like)

    @property
    def current_catalog(self) -> str:
        with self._safe_raw_sql(sg.select(sg.func("current_database"))) as cur:
            (db,) = cur.fetchone()
        return db

    @property
    def current_database(self) -> str:
        with self._safe_raw_sql(sg.select(sg.func("current_schema"))) as cur:
            (schema,) = cur.fetchone()
        return schema

    def function(self, name: str, *, database: str | None = None) -> Callable:
        n = ColGen(table="n")
        p = ColGen(table="p")
        f = self.compiler.f

        predicates = [p.proname.eq(sge.convert(name))]

        if database is not None:
            predicates.append(n.nspname.rlike(sge.convert(f"^({database})$")))

        query = (
            sg.select(
                f["pg_catalog.pg_get_function_result"](p.oid).as_("return_type"),
                f.string_to_array(
                    f["pg_catalog.pg_get_function_arguments"](p.oid), ", "
                ).as_("signature"),
            )
            .from_(sg.table("pg_proc", db="pg_catalog").as_("p"))
            .join(
                sg.table("pg_namespace", db="pg_catalog").as_("n"),
                on=n.oid.eq(p.pronamespace),
                join_type="LEFT",
            )
            .where(*predicates)
        )

        def split_name_type(arg: str) -> tuple[str, dt.DataType]:
            name, typ = arg.split(" ", 1)
            return name, self.compiler.type_mapper.from_string(typ)

        with self._safe_raw_sql(query) as cur:
            rows = cur.fetchall()

        if not rows:
            name = f"{database}.{name}" if database else name
            raise exc.MissingUDFError(name)
        elif len(rows) > 1:
            raise exc.AmbiguousUDFError(name)

        [(raw_return_type, signature)] = rows
        return_type = self.compiler.type_mapper.from_string(raw_return_type)
        signature = list(map(split_name_type, signature))

        # dummy callable
        def fake_func(*args, **kwargs): ...

        fake_func.__name__ = name
        fake_func.__signature__ = inspect.Signature(
            [
                inspect.Parameter(
                    name, kind=inspect.Parameter.POSITIONAL_OR_KEYWORD, annotation=typ
                )
                for name, typ in signature
            ],
            return_annotation=return_type,
        )
        fake_func.__annotations__ = {"return": return_type, **dict(signature)}
        op = ops.udf.scalar.builtin(fake_func, database=database)
        return op

    def get_schema(
        self,
        name: str,
        *,
        catalog: str | None = None,
        database: str | None = None,
    ):
        a = ColGen(table="a")
        c = ColGen(table="c")
        n = ColGen(table="n")

        format_type = self.compiler.f["pg_catalog.format_type"]

        # If no database is specified, assume the current database
        db = database or self.current_database

        dbs = [sge.convert(db)]

        # If a database isn't specified, then include temp tables in the
        # returned values
        if database is None and (temp_table_db := self._session_temp_db) is not None:
            dbs.append(sge.convert(temp_table_db))

        type_info = (
            sg.select(
                a.attname.as_("column_name"),
                format_type(a.atttypid, a.atttypmod).as_("data_type"),
                sg.not_(a.attnotnull).as_("nullable"),
            )
            .from_(sg.table("pg_attribute", db="pg_catalog").as_("a"))
            .join(
                sg.table("pg_class", db="pg_catalog").as_("c"),
                on=c.oid.eq(a.attrelid),
                join_type="INNER",
            )
            .join(
                sg.table("pg_namespace", db="pg_catalog").as_("n"),
                on=n.oid.eq(c.relnamespace),
                join_type="INNER",
            )
            .where(
                a.attnum > 0,
                sg.not_(a.attisdropped),
                n.nspname.isin(*dbs),
                c.relname.eq(sge.convert(name)),
            )
            .order_by(a.attnum)
        )

        type_mapper = self.compiler.type_mapper

        with self._safe_raw_sql(type_info) as cur:
            rows = cur.fetchall()

        if not rows:
            raise com.TableNotFound(name)

        return sch.Schema(
            {
                col: type_mapper.from_string(typestr, nullable=nullable)
                for col, typestr, nullable in rows
            }
        )

    def _get_schema_using_query(self, query: str) -> sch.Schema:
        name = util.gen_name(f"{self.name}_metadata")

        create_stmt = sge.Create(
            kind="VIEW",
            this=sg.table(name),
            expression=sg.parse_one(query, read=self.dialect),
            properties=sge.Properties(expressions=[sge.TemporaryProperty()]),
        )
        drop_stmt = sge.Drop(kind="VIEW", this=sg.table(name), exists=True).sql(
            self.dialect
        )

        with self._safe_raw_sql(create_stmt):
            pass
        try:
            return self.get_schema(name)
        finally:
            with self._safe_raw_sql(drop_stmt):
                pass

    def create_database(
        self, name: str, catalog: str | None = None, force: bool = False
    ) -> None:
        if catalog is not None and catalog != self.current_catalog:
            raise exc.UnsupportedOperationError(
                f"{self.name} does not support creating a database in a different catalog"
            )
        sql = sge.Create(
            kind="SCHEMA", this=sg.table(name, catalog=catalog), exists=force
        )
        with self._safe_raw_sql(sql):
            pass

    def drop_database(
        self,
        name: str,
        catalog: str | None = None,
        force: bool = False,
        cascade: bool = False,
    ) -> None:
        if catalog is not None and catalog != self.current_catalog:
            raise exc.UnsupportedOperationError(
                f"{self.name} does not support dropping a database in a different catalog"
            )

        sql = sge.Drop(
            kind="SCHEMA",
            this=sg.table(name, catalog=catalog),
            exists=force,
            cascade=cascade,
        )
        with self._safe_raw_sql(sql):
            pass

    def create_table(
        self,
        name: str,
        obj: ir.Table
        | pd.DataFrame
        | pa.Table
        | pl.DataFrame
        | pl.LazyFrame
        | None = None,
        *,
        schema: sch.SchemaLike | None = None,
        database: str | None = None,
        temp: bool = False,
        overwrite: bool = False,
    ):
        """Create a table in Postgres.

        Parameters
        ----------
        name
            Name of the table to create
        obj
            The data with which to populate the table; optional, but at least
            one of `obj` or `schema` must be specified
        schema
            The schema of the table to create; optional, but at least one of
            `obj` or `schema` must be specified
        database
            The name of the database in which to create the table; if not
            passed, the current database is used.
        temp
            Create a temporary table
        overwrite
            If `True`, replace the table if it already exists, otherwise fail
            if the table exists

        """
        if obj is None and schema is None:
            raise ValueError("Either `obj` or `schema` must be specified")
        if schema is not None:
            schema = ibis.schema(schema)

        properties = []

        if temp:
            properties.append(sge.TemporaryProperty())

        if obj is not None:
            if not isinstance(obj, ir.Expr):
                table = ibis.memtable(obj)
            else:
                table = obj

            self._run_pre_execute_hooks(table)

            query = self.compiler.to_sqlglot(table)
        else:
            query = None

        if overwrite:
            temp_name = util.gen_name(f"{self.name}_table")
        else:
            temp_name = name

        if not schema:
            schema = table.schema()

        table_expr = sg.table(temp_name, db=database, quoted=self.compiler.quoted)
        target = sge.Schema(
            this=table_expr, expressions=schema.to_sqlglot(self.dialect)
        )

        create_stmt = sge.Create(
            kind="TABLE",
            this=target,
            properties=sge.Properties(expressions=properties),
        )

        this = sg.table(name, catalog=database, quoted=self.compiler.quoted)
        with self._safe_raw_sql(create_stmt) as cur:
            if query is not None:
                insert_stmt = sge.Insert(this=table_expr, expression=query).sql(
                    self.dialect
                )
                cur.execute(insert_stmt)

            if overwrite:
                cur.execute(
                    sge.Drop(kind="TABLE", this=this, exists=True).sql(self.dialect)
                )
                cur.execute(
                    f"ALTER TABLE IF EXISTS {table_expr.sql(self.dialect)} RENAME TO {this.sql(self.dialect)}"
                )

        if schema is None:
            return self.table(name, database=database)

        # preserve the input schema if it was provided
        return ops.DatabaseTable(
            name, schema=schema, source=self, namespace=ops.Namespace(database=database)
        ).to_expr()

    def drop_table(
        self,
        name: str,
        database: str | None = None,
        force: bool = False,
    ) -> None:
        drop_stmt = sg.exp.Drop(
            kind="TABLE",
            this=sg.table(name, db=database, quoted=self.compiler.quoted),
            exists=force,
        )
        with self._safe_raw_sql(drop_stmt):
            pass

    @contextlib.contextmanager
    def _safe_raw_sql(self, *args, **kwargs):
        with contextlib.closing(self.raw_sql(*args, **kwargs)) as result:
            yield result

    def raw_sql(self, query: str | sg.Expression, **kwargs: Any) -> Any:
        import psycopg2
        import psycopg2.extras

        with contextlib.suppress(AttributeError):
            query = query.sql(dialect=self.dialect)

        con = self.con
        cursor = con.cursor()

        try:
            # try to load hstore, uuid and ipaddress extensions
            with contextlib.suppress(psycopg2.ProgrammingError):
                psycopg2.extras.register_hstore(cursor)
            with contextlib.suppress(psycopg2.ProgrammingError):
                psycopg2.extras.register_uuid(conn_or_curs=cursor)
            with contextlib.suppress(psycopg2.ProgrammingError):
                psycopg2.extras.register_ipaddress(cursor)
        except Exception:
            cursor.close()
            raise

        try:
            cursor.execute(query, **kwargs)
        except Exception:
            con.rollback()
            cursor.close()
            raise
        else:
            con.commit()
            return cursor
